/*
 * arg.c - common argument processing support functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: arg.c,v 1.6 94/05/21 09:42:32 abe Exp $";
#endif


#include "lsof.h"
#include "version.h"


/*
 * enter_network_address() - enter Internet address for searching
 */

int
enter_network_address(na)
	char *na;			/* Internet address string pointer */
{
	unsigned char *ap;
	int at = 0;
	struct hostent *he;
	char *hn = NULL;
	MALLOC_S l;
	struct nwad *n = NULL;
	struct nwad *nn;
	char *p, *wa;
	struct servent *se;

/*
 * Allocate and initialize network address structure.
 */
	if ((n = (struct nwad *)malloc(sizeof(struct nwad))) == NULL) {

no_nwad_space:

	    (void) fprintf(stderr,
		"%s: no space for network address from \"%s\"\n", Pn, na);
	    exit(1);
	}
	n->proto = NULL;
	n->a[0] = n->a[1] = n->a[2] = n->a[3] = 0;
	n->port = -1;
/*
 * Process protocol name, optionally followed by a '@' and a host name or
 * Internet address, or a ':' and a service name or port number.
 */
	wa = na;
	if (*wa && *wa != '@' && *wa != ':') {
	    for (p = wa; *wa && *wa != '@' && *wa != ':'; wa++)
		;
	    if ((l = wa - p)) {
		if ((n->proto = (char *)malloc(l + 1)) == NULL) {
		    (void) fprintf(stderr,
			"%s: no space for protocol name from \"%s\"\n",
			Pn, na);
nwad_exit:
		    if (n->proto)
			(void) free((FREE_P *)n->proto);
		    if (hn)
			(void) free((FREE_P *)hn);
		    (void) free((FREE_P *)n);
		    return(1);
		}
		(void) strncpy(n->proto, p, l);
		n->proto[l] = '\0';
	    /*
	     * Convert protocol name to lower case.
	     */
		for (p = n->proto; *p; p++) {
		    if (*p >= 'A' && *p <= 'Z')
			*p = *p - 'A' + 'a';
		}
	    }
	}
/*
 * Process an Internet address (1.2.3.4) or host name, preceded by a '@'
 * and optionally followed by a colon and a service name or port number.
 */
	if (*wa == '@') {
	    wa++;
	    if ( ! *wa || *wa == ':') {

unacc_address:
		(void) fprintf(stderr,
		    "%s: unacceptable Internet address in \"%s\"\n", Pn, na);
		goto nwad_exit;
	    }
	    if (*wa < '0' || *wa > '9') {
	/*
	 * Assemble host name.
	 */
		for (p = wa; *p && *p != ':'; p++)
		    ;
		if ((l = p - wa)) {
		    if ((hn = (char *)malloc(l + 1)) == NULL) {
			fprintf(stderr,
			    "%s: no space for host name from \"%s\"\n", Pn, na);
			goto nwad_exit;
		    }
		    (void) strncpy(hn, wa, l);
		    hn[l] = '\0';
		    if ((he = gethostbyname(hn)) == NULL) {
			fprintf(stderr, "%s: unknown host name in \"%s\"\n",
			    Pn, na);
			goto nwad_exit;
		    }
		    ap = (unsigned char *)he->h_addr;
		    n->a[0] = *ap++;
		    n->a[1] = *ap++;
		    n->a[2] = *ap++;
		    n->a[3] = *ap;
		    at = 1;
		}
		wa = p;
	    } else {
	/*
	 * Assemble Internet address.
	 */
		for (l = 0; *wa; wa++) {
		    if (*wa == ':')
			break;
		    if (*wa == '.') {
			l++;
			if (l > 3)
			    break;
			continue;
		    }
		    if (*wa < '0' || *wa > '9')
			goto unacc_address;
		    n->a[l] = (10 * n->a[l]) + *wa - '0';
		    if (n->a[l] > 255)
			goto unacc_address;
		}
		if (l != 3 || (n->a[0] == 0 && n->a[1] == 0 && n->a[2] == 0
		&& n->a[3] == 0))
		    goto unacc_address;
	    }
	}
/*
 * Process a service name or port number, preceded by a colon.
 */
	if (*wa && (*wa != ':' || ! *(++wa))) {

unacc_port:
	    (void) fprintf(stderr, "%s: unacceptable port number in \"%s\"\n",
		Pn, na);
	    goto nwad_exit;
	}
	if (*wa && (*wa < '0' || *wa > '9')) {

	/*
	 * Convert service name to port number, using already-specified
	 * protocol name.
	 */
	    if (! n->proto) {
		(void) fprintf(stderr, "%s: must specify protocol in \"%s\"\n",
		    Pn, na);
		goto nwad_exit;
	    }
	    if ((se = getservbyname(wa, n->proto)) == NULL) {
		(void) fprintf(stderr,
		    "%s: unknown service \"%s\" for protocol %s in \"%s\"\n",
		    Pn, wa, n->proto, na);
		goto nwad_exit;
	    }
	    n->port = (int)se->s_port;
	} else if (*wa) {

	/*
	 * Assemble port number.
	 */
	    for (n->port = 0; *wa; wa++) {
		if (*wa < '0' || *wa > '9')
		    goto unacc_port;
		n->port = (n->port * 10) + *wa - '0';
	    }
	}
/*
 * Test completed specification -- it must contain at least one of: protocol,
 * Internet address or port.  If correct, link into search list.
 */
	if ( ! n->proto
	&&  n->a[0] == 0 && n->a[1] == 0 && n->a[2] == 0 && n->a[3] == 0
	&&  n->port == -1) {
	    (void) fprintf(stderr,
		"%s: no Internet address specified in \"%s\"\n",
		Pn, na);
	    goto nwad_exit;
	}
	n->next = Nwad;
	Nwad = n;
	if (at) {

	/*
	 * If the network address came from gethostbyname(), enter all
	 * the addresses for the name.
	 */
	    for (; he->h_addr_list[at]; at++) {
		if ((nn = (struct nwad *)malloc(sizeof(struct nwad))) == NULL)
		    goto no_nwad_space;
		*nn = *n;
		ap = (unsigned char *)he->h_addr_list[at];
		nn->a[0] = *ap++;
		nn->a[1] = *ap++;
		nn->a[2] = *ap++;
		nn->a[3] = *ap;
		nn->next = Nwad;
		Nwad = nn;
	    }
	}
	return(0);
}


/*
 * enter_pid() - enter Process IDentifier for searching
 */

int
enter_pid(p)
	char *p;			/* PID string pointer */
{
	int i, pid;
	char *s;

	for (s = p; *s;) {
	/*
	 * Assemble Process IDentifier.
	 */
		for (pid = 0; *s && *s != ','; *s++) {

#if	defined(__STDC__)
			if ( ! isdigit(*s))
#else
			if ( ! isascii(*s) || ! isdigit(*s))
#endif	/* __STDC__ */

			{
				(void) fprintf(stderr,
					"%s: illegal PID specification: %s\n",
					Pn, p);
				return(1);
			}
			pid = (pid * 10) + *s - '0';
		}
		if (*s)
			s++;
	/*
	 * Avoid entering duplicates.
	 */
		for (i = 0; i < Npid; i++) {
			if (pid == Spid[i])
				break;
		}
		if (i < Npid)
			continue;
	/*
	 * Allocate PID table space.
	 */
		if (Npid >= Mxpid) {
			Mxpid += PIDINCR;
			if (Spid == NULL)
				Spid = (int *)malloc((MALLOC_S)
					(sizeof(int *) * Mxpid));
			else
				Spid = (int *)realloc((MALLOC_P *)Spid,
					(MALLOC_S)(sizeof(int *) * Mxpid));
			if (Spid == NULL) {
				(void) fprintf(stderr, "%s: no space for PIDs",
					Pn);
				exit(1);
			}
		}
		Spid[Npid++] = pid;
		Npuns = Npid;
	}
	return(0);
}


/*
 * enter_uid() - enter User Identifier for searching
 */

int
enter_uid(u)
	char *u;			/* User IDentifier string pointer */
{
	int err, i, nn;
	char lnm[LOGINML];
	struct passwd *pw;
	char *s;
	uid_t uid;

	for (err = 0, s = u; *s;) {
	/*
	 * Assemble next User IDentifier.
	 */
		for (i = nn = uid = 0; *s && *s != ','; i++, s++) {
			if (i >= LOGINML-1) {
				(void) fprintf(stderr,
					"%s: illegal UID in %s\n", Pn, u);
				return(1);
			}
			lnm[i] = *s;
			if (nn)
				continue;

#if	defined(__STDC__)
			if (isdigit(*s))
#else
			if (isascii(*s) && isdigit(*s))
#endif	/* __STDC__ */

				uid = (uid * 10) + *s - '0';
			else
				nn++;
		}
		if (*s)
			s++;
		if (nn) {
		       lnm[i] = '\0';

#if	defined(HASPWSTAYOPEN)
			pw_stay_open();
#endif

			if ((pw = getpwnam(lnm)) == NULL) {
				(void) fprintf(stderr,
					"%s: can't get UID for %s\n",
					Pn, lnm);
				err = 1;
				continue;
			} else
				uid = pw->pw_uid;
		}

#if	defined(HASSECURITY)
	/*
	 * If the security mode is enabled, only the root user may list files
	 * belonging to user IDs other than the real user ID of this lsof
	 * process.
	 */
		if (Myuid && uid != Myuid) {
		    (void) fprintf(stderr,
		    "%s: ID %d request rejected because of security mode.\n",
			Pn, uid);
		    err = 1;
		    continue;
		}
#endif	/* HASSECURITY */

	/*
	 * Avoid entering duplicates.
	 */
		for (i = 0; i < Nuid; i++) {
			if (uid == Suid[i])
				break;
		}
		if (i < Nuid)
			continue;
	/*
	 * Allocate space for User IDentifier.
	 */
		if (Nuid >= Mxuid) {
			Mxuid += UIDINCR;
			if (Suid == NULL)
				Suid = (uid_t *)malloc((MALLOC_S)
					(sizeof(uid_t) * Mxuid));
			else
				Suid = (uid_t *)realloc((MALLOC_P *)Suid,
					(MALLOC_S)(sizeof(uid_t) * Mxuid));
			if (Suid == NULL) {
				(void) fprintf(stderr, "%s: no space for UIDs",
					Pn);
				exit(1);
			}
		}
		Suid[Nuid++] = uid;
	}
	return(err);
}


/*
 * usage() - display usage and exit
 */

void
usage(xv)
	int xv;				/* exit value */
{
	(void) fprintf(stderr, "%s %s usage: [-adhHlmnNoOsPt]", Pn, VERSION);
	(void) fprintf(stderr, " [-c c] [-i i]");

#if	defined(HASKOPT)
	(void) fprintf(stderr, " [-k k]");
#endif

	(void) fprintf(stderr, " [-p l] [-S t] [-u l]\n  [-U] [names]\n");
	(void) fprintf(stderr, "Defaults are enclosed in parentheses.");
	(void) fprintf(stderr, "  Separate items in a set by commas.\n");
	(void) fprintf(stderr, "  %-37.37s",
		"-a     AND selections (OR)");
	(void) fprintf(stderr,
		"-c c   kernel memory path (%s)\n", KMEM);
	(void) fprintf(stderr, "  %-37.37s",
		"-d     enable /dev error messages");
	(void) fprintf(stderr, "-h     list help\n");
	(void) fprintf(stderr,
		"  -H     don't list host names\n");
	(void) fprintf(stderr,
		"  -i i   select by Internet address:");
	(void) fprintf(stderr,
		" [protocol][@name|number][:service|port]\n");

#if	defined(HASKOPT)
	(void) fprintf(stderr,
		"  -k k   kernel symbols path (%s)\n", N_UNIX);
#endif

	(void) fprintf(stderr, "  %-37.37s",
		"-l     list UID numbers");
	(void) fprintf(stderr,
		"-m     stop on mount table errors\n");
	(void) fprintf(stderr, "  %-37.37s",
		"-n     select Internet files");
	(void) fprintf(stderr,
		"-N     select NFS files\n");
	(void) fprintf(stderr, "  %-37.37s",
		"-p l   select by PID set");
	(void) fprintf(stderr,
		"-P     don't list port names\n");
	(void) fprintf(stderr, "  %-37.37s",
		"-o     always list file offset");
	(void) fprintf(stderr,
		"-O     don't order by PID\n");
	(void) fprintf(stderr, "  %-37.37s",
		"-s     always list file size");
	(void) fprintf(stderr,
		"-S t   readlink/stat timeout sec (%d)\n", TMLIMIT);
	(void) fprintf(stderr, "  %-37.37s",
		"-t     terse listing");
	(void) fprintf(stderr,
		"-u l   select by user login/UID set\n");
	(void) fprintf(stderr, "  %s\n",
		"-U     select Unix files");
	(void) fprintf(stderr,
		"  names  select named files or files on named file systems\n");
	(void) fprintf(stderr,
		"The security mode is %s.\n",

#if	defined(HASSECURITY)
		"enabled -- only the root user may list all open files"
#else
		"disabled -- anyone may list all open files"
#endif

		);
	exit(xv);
}
