/*
 * ddev.c - AIX device support functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: ddev.c,v 1.2 94/05/06 15:59:24 abe Exp $";
#endif


#include "lsof.h"

_PROTOTYPE(static void readdev,(void));
_PROTOTYPE(static void stkdir,(char *p));


/*
 * Local static values
 */

int Dn = 0;				/* directory stack entries allocated */
char **Dstk = NULL;			/* directory stack */
int Dx = 0;				/* directory stack index */


/*
 * getchan() - get channel from file path name
 */

int
getchan(p)
	char *p;			/* file path name */
{
	int ch;
	char *s;

	if ((s = strrchr(p, '/')) == NULL) 
		return(-1);
	if (*(++s) == '\0')
		return(-1);
	for (ch = 0; *s; s++) {

#if	defined(__STDC__)
		if ( ! isdigit(*s))
#else
		if ( ! isascii(*s) || ! isdigit(*s))
#endif	/* __STDC__ */

			return(-1);
		ch = (ch * 10) + *s - '0';
	}
	return(ch);
}


/*
 * printchdevname() - print character device name
 */

int
printchdevname(rdev, f)
	int rdev;			/* device */
	int f;				/* 1 = follow with '\n' */
{
	int low, hi, mid;
	struct dev *dp;

	if (!Sdev)
		readdev();
/*
 * Search device table for match.
 */
	low = mid = 0;
	hi = Ndev - 1;
        while (low <= hi) {
                mid = (low + hi) / 2;
		dp = Sdev[mid];
		if ((long)rdev < (long)dp->rdev)
			hi = mid - 1;
		else if ((long)rdev > (long)dp->rdev)
                        low = mid + 1;
                else {
			if (Lf->inode < 0L || (ino_t)Lf->inode == dp->inode) {
				if (Lf->ch < 0) {
				    if (f)
					(void) puts(dp->name);
				    else
					(void) fputs(dp->name, stdout);
				} else {
				    (void) printf("%s/%d", dp->name, Lf->ch);
				    if (f)
					putchar('\n');
				}
				return(1);
			}
			if (Lf->inode < (long)dp->inode)
				hi = mid - 1;
			else
				low = mid + 1;
		}
        }
	return(0);
}


/*
 * readdev() - read names, modes and device types of everything in /dev
 */

void
readdev()
{
	DIR *dfp;
	struct dirent *dp;
	int err = 0;
	static int first = 1;
	int i = 0;
	MALLOC_S nl;
	char path[MAXNAMLEN+1];
	int pl;
	struct stat sb;

	if (!first)
		return;
	first = 0;
	(void) stkdir("/dev");
/*
 * Unstack the next /dev or /dev/<subdirectory> directory.
 */
	while (--Dx >= 0) {
		(void) strcpy(path, Dstk[Dx]);
		if ((dfp = opendir(path)) == NULL) {
			if (Fdevmsg)
				(void) fprintf(stderr, "%s: can't open %s\n",
					Pn, path);
			continue;
		}
		(void) strcat(path, "/");
		pl = strlen(path);
		(void) free((FREE_P *)Dstk[Dx]);
		Dstk[Dx] = NULL;
	/*
	 * Scan the directory.
	 */
		for (dp = readdir(dfp); dp; dp = readdir(dfp)) {
			if (dp->d_ino == 0 || dp->d_name[0] == '.')
				continue;
		/*
		 * Form the full path name and get its status.
		 */
			if ((nl = pl + dp->d_namlen) >= sizeof(path)) {
				(void) fprintf(stderr,
					"%s: /dev entry name too long: %s\n",
					Pn, dp->d_name);
				exit(1);
			}
			(void) strncpy(&path[pl], dp->d_name,
				(STRNCPY_L)dp->d_namlen);
			path[nl++] = '\0';
			if (lstat(path, &sb) != 0) {
				(void) fprintf(stderr,
					"%s: can't lstat %s: %s\n",
					Pn, path, strerror(errno));
				err = 1;
				continue;
			}
		/*
		 * If it's a subdirectory, stack its name for later
		 * processing.
		 */
			if ((sb.st_mode & S_IFMT) == S_IFDIR) {
				(void) stkdir(path);
				continue;
			}
		/*
		 * Skip all but character devices.
		 */
			if ((sb.st_mode & S_IFMT) != S_IFCHR)
				continue;
		/*
		 * Make room for another Devtp[] entry.
		 */
			if (i >= Ndev) {
				Ndev += DEVINCR;
				if (Devtp == NULL)
				    Devtp = (struct dev *)malloc(
					(MALLOC_S)(sizeof(struct dev) * Ndev));
				else
				    Devtp = (struct dev *)realloc(
					(MALLOC_P *)Devtp,
					(MALLOC_S)(sizeof(struct dev)*Ndev));
				if (Devtp == NULL) {
					(void) fprintf(stderr,
						"%s: no space for devices\n",
						Pn);
					exit(1);
				}
			}
		/*
		 * Store the device number, inode number, and name in
		 * the Devtp[] entry.
		 */
			Devtp[i].rdev = sb.st_rdev;
			Devtp[i].inode = sb.st_ino;
			if ((Devtp[i].name = (char *)malloc(nl)) == NULL) {
				(void) fprintf(stderr,
					"%s: no space for /dev/%s\n",
					Pn, dp->d_name);
				exit(1);
			}
			(void) strcpy(Devtp[i].name, path);
			i++;
		}
		(void) closedir(dfp);
	}
/*
 * Free any directory stack space.
 */
	if (Dstk != NULL)
		(void) free((FREE_P *)Dstk);
/*
 * Reduce the Devtp[] table to its minimum size.
 */
	if (Ndev > i) {
		Ndev = i;
		Devtp = (struct dev *)realloc((MALLOC_P *)Devtp,
		      (MALLOC_S)(sizeof(struct dev) * Ndev));
	}
	if (err)
		exit(1);
/*
 * Allocate sorting pointers and sort them by Devtp[] device number.
 */
	if ((Sdev = (struct dev **)malloc((MALLOC_S)(sizeof(struct dev *)
	    * Ndev)))
	== NULL) {
		(void) fprintf(stderr, "%s: no space for device pointers\n",
			Pn);
		exit(1);
	}
	for (i = 0; i < Ndev; i++) {
		Sdev[i] = &Devtp[i];
	}
	(void) qsort((QSORT_P *)Sdev,(size_t)Ndev,(size_t)sizeof(struct dev *),
		compdev);
}


/*
 * stkdir() - stack directory name
 */

static void
stkdir(p)
	char *p;		/* directory path */
{
	if (Dstk == NULL) {

	/*
	 * Allocate first entry.
	 */
		if ((Dstk = (char **)malloc(sizeof(char *))) == NULL) {

stkdir_nospace:

			(void) fprintf(stderr,
				"%s: no space for directory stack at %s\n",
				Pn, p);
			exit(1);
		}
		Dn = 1;
		Dx = 0;
	} else if (Dx >= Dn) {

	/*
	 * Allocate additional space as required.
	 */
		Dn++;
		if ((Dstk = (char **)realloc((MALLOC_P *)Dstk,
		          (MALLOC_S)(Dn * sizeof(char *))))
		== NULL)
			goto stkdir_nospace;
	}
/*
 * Allocate space for the name, copy it there and put its pointer on the stack.
 */
	if ((Dstk[Dx] = (char *)malloc((MALLOC_S)(strlen(p) + 1))) == NULL) {
		(void) fprintf(stderr, "%s: no space for %s\n", Pn, p);
		exit(1);
	}
	(void) strcpy(Dstk[Dx], p);
	Dx++;
}


/*
 * The find_ch_ino() function is obtained from ../common/fchi.frag.
 */
