

/*
 * The caller must define CKFA_EXPDEV and an expdev() macro if stat()
 * buffer device numbers need to be defined.  (This is an EP/IX artifact.)
 *
 * The caller must define CKFA_MPXCHAN to enable the saving of
 * a multiplexed channel number from the path name.  (This is an AIX
 * artifact.)
 */


/*
 * ck_file_arg() - check file arguments
 */

int
ck_file_arg(i, ac, av)
	int i;			/* first file argument index */
	int ac;			/* argument count */
	char *av[];		/* argument vector */
{
	int err = 0;
	char *fnm, *fsnm, *path;
	int ftype, j;
	struct mounts *mp;
	struct stat sb;
	struct sfile *sfp;

#if	defined(CKFA_EXPDEV)
	dev_t dev, rdev;
#endif

	for (; i < ac; i++) {
		if ((path = Readlink(av[i])) == NULL) {
			err = 1;
			continue;
		}
	/*
	 * Remove extra terminating `/'.
	 *
	 * Check for file system argument.
	 */
		if ((j = strlen(path)) > 1 && path[j-1] == '/')
			path[j-1] = '\0';
		for (ftype = 1, mp = Mtab; mp; mp = mp->next) {
			if (strcmp(mp->dir, path) == 0) {
				ftype = 0;
				fnm = path;
				fsnm = mp->fsname;
				break;
			}
			if (strcmp(mp->fsname, path) == 0) {
				ftype = 0;
				fnm = mp->dir;
				fsnm = path;
				break;
			}
		}
		if (ftype) {
			fnm = path;
			fsnm = NULL;
		} else {

#if	defined(CKFA_EXPDEV)
			dev = mp->dev;
			rdev = mp->rdev;
#else
			sb.st_dev = mp->dev;
			sb.st_rdev = mp->rdev;
#endif

			sb.st_ino = mp->inode;
			sb.st_mode = mp->mode;
		}
	/*
	 * Stat the argument to obtain its mode and device.
	 */
		if (ftype && statsafely(fnm, &sb) != 0) {
			(void) fprintf(stderr, "%s: status error on %s: %s\n",
				Pn, fnm, strerror(errno));
			err = 1;
			continue;
		}

#if	defined(CKFA_EXPDEV)
	/*
	 * Expand device numbers before saving, so that they match the
	 * already-expanded Mtab device numbers.  (This is an EP/IX
	 * 2.1.1 and above artifact.)
	 */
		if (ftype) {
			dev = expdev(sb.st_dev);
			rdev = expdev(sb.st_rdev);
		}
#endif

	/*
	 * Allocate an sfile structure and fill in the type, inode,
	 * find-flag and linkages.
	 */
		if ((sfp = (struct sfile *)malloc(sizeof(struct sfile)))
		== NULL) {
			(void) fprintf(stderr, "%s: no space for files\n", Pn);
			exit(1);
		}
		sfp->next = Sfile;
		Sfile = sfp;
		sfp->type = ftype;
		sfp->i = sb.st_ino;
		sfp->f = 0;
	/*
	 * Store the file name and file system name pointers in the sfile
	 * structure, allocating space as necessary.
	 */
		if (fnm == NULL || fnm == path)
			sfp->name = fnm;
		else {
		    if ((sfp->name = (char *)malloc((MALLOC_S)(strlen(fnm)+1)))
		    == NULL) {
			(void) fprintf(stderr,
			    "%s: no space for file name %s\n", Pn, fnm);
			exit(1);
		    }
		    (void) strcpy(sfp->name, fnm);
		}
		if (fsnm == NULL || fsnm == path)
			sfp->devnm = fsnm;
		else {
		    if ((sfp->devnm=(char *)malloc((MALLOC_S)(strlen(fsnm)+1)))
		    == NULL) {
			(void) fprintf(stderr,
			    "%s: no space for file system name %s\n", Pn, fsnm);
			exit(1);
		    }
		    (void) strcpy(sfp->devnm, fsnm);
		}
		if ((sfp->aname = (char *)malloc((MALLOC_S)(strlen(av[i]) + 1)))
		== NULL) {
			(void) fprintf(stderr,
			    "%s: no space for argument file name %s\n",
			    Pn, av[i]);
			exit(1);
		}
		(void) strcpy(sfp->aname, av[i]);
	/*
	 * Save the stat() buffer mode value in the sfile structure.
	 * Use st_rdev if the mode value is S_IFBLK or S_IFCHR; otherwise
	 * use st_dev.
	 */
		sfp->mode = sb.st_mode & S_IFMT;
		if (sfp->mode == S_IFBLK || sfp->mode == S_IFCHR)

#if	defined(CKFA_EXPDEV)
			sfp->dev = rdev;
		else
			sfp->dev = dev;
#else
			sfp->dev = sb.st_rdev;
		else
			sfp->dev = sb.st_dev;
#endif

#if	defined(CKFA_MPXCHAN)
	/*
	 * Save a (possible) multiplexed channel number.  (This is an AIX
	 * artifact.)
	 */
		sfp->ch = getchan(path);
#endif

	}
	return(err);
}
