/*
 * dmnt.c - DYNIX mount support functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: dmnt.c,v 1.2 94/04/22 10:27:41 abe Exp $";
#endif


#include "lsof.h"


/*
 * readvfs() - read vfs structure
 */

struct l_vfs *
readvfs(ka)
	struct vfs *ka;			/* vfs structure kernel address */
{
	struct mount m;
	struct mntinfo mi;
	struct vfs tv;
	struct l_vfs *vp;

	if (!ka)
		return(NULL);
	for (vp = Lvfs; vp; vp = vp->next) {
		if (ka == vp->addr)
			return(vp);
	}
	if ((vp = (struct l_vfs *)malloc(sizeof(struct l_vfs))) == NULL) {
		(void) fprintf(stderr, "%s: PID %d, no space for vfs\n",
			Pn, Lp->pid);
		exit(1);
	}
	vp->dev = 0;
	vp->dir = NULL;
	vp->fsname = NULL;
	if (kread((KA_T)ka, (char *)&tv, sizeof(tv))) {
		(void) free((FREE_P *)vp);
		return(NULL);
	}
/*
 * Complete mount information.
 */
	if (Ntype == N_NFS) {

	/*
	 * DYNIX NFS vnode device values have a major number of 255.
	 * The minor number is a serial number found in the mntinfo
	 * structure to which the virtual file system's vfs_data points.
	 */ 
		if (tv.vfs_data
		&&  kread((KA_T)tv.vfs_data, (char *)&mi, sizeof(mi)) == 0)
			(void) completevfs(vp, makedev(255, (int)mi.mi_mntno));
	} else {
		if (kread((KA_T)tv.vfs_data, (char *)&m, sizeof(m)) == 0)
			(void) completevfs(vp, (int)m.m_dev);
	}
	vp->next = Lvfs;
	vp->addr = ka;
	Lvfs = vp;
	return(vp);
}


/*
 * The readmnt() function is obtained from ../common/rmnt.frag.
 * The completevfs() function is obtained from ../common/cvfs.frag.
 */

#define	CVFS_DEVSAVE 	1
#define	EXPDEV(n)	n
/* MNTSKIP is not defined. */
#define	USECVFS		1
