/*
 * dproc.c - EP/IX process access functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: dproc.c,v 1.3 94/05/27 12:56:56 abe Exp $";
#endif

#include "lsof.h"


/*
 * Local static values
 */

static KA_T Kp;				/* kernel's process table address */
static int Knp;				/* number of processes according
					 * to kernel */
static int Np;				/* number of processes read from
					 * kernel */

#if	_EPIXV>=20101
static struct proc *P;			/* local proc structure table */
static KA_T *Pa;			/* kernel address for each P[] entry */
#endif


_PROTOTYPE(static void get_kernel_access,(void));
_PROTOTYPE(static int mread,(KA_T addr, char *buf, READLEN_T len));

#if	_EPIXV>=20101
_PROTOTYPE(static void read_proc,(void));
#endif


/*
 * gather_proc_info() -- gather process information
 */

void
gather_proc_info()
{
	int err, i, pid, ps, px;
	struct proc *p;
	int p2bsh, p2psh;

#if	_EPIXV<20101
	KA_T pa;
	struct proc pbuf;
#else
	struct cred pcred;
	struct pid pids;
#endif

	short pss, sf;
	struct user *u;
	char *ua, *ub, *ubl;
	MALLOC_S ual;
	struct file **uf;
	uid_t uid;
/*
 * Allocate user area buffer.
 */
	ual = (MALLOC_S) (sizeof(struct user) 
	    + ((NOFILES_UAREA - 1) * sizeof(struct file *)));
	if ((ua = (char *)malloc(ual)) == NULL) {
		(void) fprintf(stderr, "%s: no space for %d byte user area\n",
			Pn, ual);
		exit(1);
	}
	u = (struct user *)ua;
	ubl = ua + ual;
/*
 * Define user area paging parameters.
 */
	if ((ps = getpagesize()) == 4096)
		p2bsh = p2psh = 12;
	else if (ps == 16384) {
		p2bsh = 14;
		p2psh = 10;
	} else {
		(void) fprintf(stderr, "%s: unknown page size (%d)\n", Pn, ps);
		exit(1);
	}
/*
 * Examine proc structures and their associated information.
 */
#if	_EPIXV<20101
	p = &pbuf;
#endif

	for (px = 0; px < Np; px++) {

#if	_EPIXV<20101
	/*
	 * Read the proc structure.  Reject uninteresting ones.
	 *
	 * Set the Process ID and User ID.
	 */
		pa = (KA_T)(Kp + (long)(px * sizeof(struct proc)));
		if (kread(pa, (char *)&pbuf, sizeof(pbuf)))
			continue;
		if (p->p_stat == 0 || p->p_stat == SZOMB)
			continue;
		pid = p->p_pid;
		uid = (uid_t)p->p_uid;
#else
	/*
	 * Set the proc structure addresses.
	 */
		p = &P[px];
		Kp = Pa[px];
	/*
	 * Get Process ID and User ID.
 	 */
		if (p->p_pidp == NULL
		||  kread((KA_T)p->p_pidp, (char *)&pids, sizeof(pids)))
			continue;
		pid = pids.pid_id;
		if (p->p_cred == NULL
		||  kread((KA_T)p->p_cred, (char *)&pcred, sizeof(pcred)))
			continue;
		uid = (uid_t)pcred.cr_ruid;
#endif

	/*
	 * See if process is excluded.
	 */
		if (is_proc_excl(pid, uid, &pss, &sf))
			continue;
	/*
	 * Read the user area.
	 */
		for (err = i = 0, ub = ua; ub < ubl; i++, ub += ps) {
			if (mread((KA_T)((p->p_ubptbl[i].pgi.pg_pde >> p2psh)
			   << p2bsh), ub,
			   (READLEN_T)(((ubl - ub) > ps) ? ps : (ubl - ub))))
			{
				err++;
				break;
			}
		}
		if (err || (KA_T)u->u_procp != Kp)
			continue;
	/*
	 * Allocate a local process structure.
	 */
		alloc_lproc(pid, uid, u->u_comm, pss, sf);
		Plf = NULL;
	/*
	 * Save current working directory information.
	 */
		if (u->u_cdir) {
			alloc_lfile(CWD, -1);
			process_node((caddr_t)u->u_cdir);
			if (Lf->sf)
				link_lfile();
		}
	/*
	 * Save root directory information.
	 */
		if (u->u_rdir) {
			alloc_lfile(RTD, -1);
			process_node((caddr_t)u->u_rdir);
			if (Lf->sf)
				link_lfile();
		}
	/*
	 * Print information on the text file.
	 */
		if (u->u_exdata.vp != NULL) {
			alloc_lfile(" txt", -1);
			process_node((caddr_t)u->u_exdata.vp);
			if (Lf->sf)
				link_lfile();
		}
	/*
	 * Loop through user's files.
	 */
		uf = (struct file **)&u->u_ofile;
		for (i = 0; i < u->u_nofiles; i++) {
			if (uf[i+1]) {
                                alloc_lfile(NULL, i);
				process_file(uf[i+1]);
				if (Lf->sf)
					link_lfile();
			}
		}
	/*
	 * Examine results.
	 */
		if (examine_lproc())
			return;
	}
}


/*
 * get_kernel_access() - access the required information in the kernel
 */

static void
get_kernel_access()
{

/*
 * Open kernel memory accesses.
 */
	if ((Kmem = open(Core ? Core : KMEM, O_RDONLY, 0)) < 0) {
		(void) fprintf(stderr, "%s: can't open %s: %s\n", Pn,
			Core ? Core : KMEM, strerror(errno));
		exit(1);
	}
	if ((Mem = open("/dev/mem", O_RDONLY, 0)) < 0) {
		(void) fprintf(stderr, "%s: can't open /dev/mem: %s\n",
			Pn, strerror(errno));
		exit(1);
	}
/*
 * Access kernel symbols.
 */
	if (nlist(Nmlst ? Nmlst : N_UNIX, Nl) < 0) {
                (void) fprintf(stderr, "%s: can't read namelist from %s\n",
			Pn, Nmlst ? Nmlst : N_UNIX);
                exit(1);
	}

#if	_EPIXV>=20101
	Vnfops = (struct fileops *) Nl[X_VN_FOPS].n_value;
	if (Nl[X_PROC].n_value == NULL
	||  kread((KA_T)Nl[X_PROC].n_value, (char *)&Kp, sizeof(Kp))
	||  Kp == NULL)
#else
	Kp = (KA_)Nl[X_PROC].n_value;
	if (Kp == NULL)
#endif

	{
		(void) fprintf(stderr, "%s: can't read proc table info\n", Pn);
		exit(1);
	}
	if (Nl[X_NPROC].n_value == NULL
	||  kread((KA_T)Nl[X_NPROC].n_value, (char *)&Knp, sizeof(Knp))
	||  Knp < 1)
		Knp = PROCDFLT;

#if	_EPIXV>=20101
	(void) read_proc();
#endif

}


/*
 * initialize() - perform all initialization
 */

void
initialize()
{
	get_kernel_access();
	(void) iuidcache(Np);
}


/*
 * kread() - read from kernel memory
 */

int
kread(addr, buf, len)
	KA_T addr;			/* kernel memory address */
	char *buf;			/* buffer to receive data */
	READLEN_T len;			/* length to read */
{
	int br;

	if (lseek(Kmem, addr, L_SET) == (off_t)-1L)
		return(-1);
	br = read(Kmem, buf, len);
	return((br == len) ? 0 : 1);
}


/*
 * mread() -- read from /dev/mem
 */

static int
mread(addr, buf, len)
	KA_T addr;			/* /dev/mem address */
	char *buf;			/* buffer to receive data */
	READLEN_T len;			/* length to read */
{
	int br;

	if (lseek(Mem, addr, L_SET) == (off_t)-1L)
		return(1);
	br = read(Mem, buf, len);
	return((br == len) ? 0 : 1);
}


#if	_EPIXV>=20101
/*
 * read_proc() - read proc structures
 */

static void
read_proc()
{
	int inc, sz, try;
	KA_T kp;
	struct proc *p;

/*
 * Try PROCTRYLM times to read a valid proc table.
 */
	for (sz = Knp + PROCDFLT/4, inc = PROCDFLT/2, try = 0;
	     try < PROCTRYLM; try++)
	{

	/*
	 * Pre-allocate space for sz proc structures.
	 */
		if ((P = (struct proc *)malloc(
			(MALLOC_S)(sizeof(struct proc) * sz)))
		== NULL) {
			(void) fprintf(stderr, "%s: no proc table space\n",
				Pn);
			exit(1);
		}
		if ((Pa = (KA_T *)malloc((MALLOC_S)(sizeof(KA_T) * sz)))
		== NULL) {
			(void) fprintf(stderr, "%s: no proc pointer space\n",
				Pn);
			exit(1);
		}
	/*
	 * Accumulate proc structures.
	 */
		for (kp = Kp, Np = 0; kp; ) {
			if (kread(kp, (char *)&P[Np], sizeof(struct proc))) {
				Np = 0;
				break;
			}
			Pa[Np] = kp;
			kp = (KA_T)P[Np].p_next;
			if (P[Np].p_stat == 0 || P[Np].p_stat == SZOMB)
				continue;
			Np++;
			if (Np >= sz) {

			/*
			 * Expand the local proc table.
			 */
				sz += inc;
				if ((P = (struct proc *)realloc((MALLOC_P *)P,
					(MALLOC_S)(sizeof(struct proc) * sz)))
				== NULL) {
					(void) fprintf(stderr,
						"%s: no more (%d) proc space\n",
						Pn, sz);
					exit(1);
				}
				if ((Pa = (KA_T *)realloc((MALLOC_P *)Pa,
					(MALLOC_S)(sizeof(KA_T) * sz)))
				== NULL) {
					(void) fprintf(stderr,
					    "%s: no more (%d) proc ptr space\n",
					    Pn, sz);
					exit(1);
				}
			}
		}
	/*
	 * If not enough processes were saved in the local table, prepare
	 * to try again.
	 */
		if (Np >= PROCMIN)
			break;
		(void) free((FREE_P *)P);
		P = NULL;
		(void) free((FREE_P *)Pa);
		Pa = NULL;
	}
/*
 * Quit if no proc structures were stored in the local table.
 */
	if (try >= PROCTRYLM) {
		(void) fprintf(stderr, "%s: can't read proc table\n", Pn);
		exit(1);
	}
	if (Np < sz) {

	/*
	 * Reduce the local proc structure table size to a minimum.
	 */
		if ((P = (struct proc *)realloc(P,
			 (MALLOC_S)(sizeof(struct proc) * Np)))
		== NULL) {
			(void) fprintf(stderr,
				"%s: can't reduce proc table to %d\n",
				Pn, Np);
			exit(1);
		}
		if ((Pa = (KA_T *)realloc((MALLOC_P *)Pa,
			  (MALLOC_S)(sizeof(KA_T) * Np)))
		== NULL) {
			(void) fprintf(stderr,
				"%s: can't reduce proc ptrs to %d\n",
				Pn, Np);
			exit(1);
		}
	}
}
#endif	/* _EPIXV >= 20101 */
