/*
 * dlsof.h - DEC OSF/1 header file for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */


#if	!defined(DECOSF_LSOF_H)
#define	DECOSF_LSOF_H	1

#include <fcntl.h>
#include <fstab.h>
#include <sys/mount.h>
#include <dirent.h>
#include <nlist.h>
#include <setjmp.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <cdfs/cdfsnode.h>
#include <rpc/types.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/table.h>
#include <sys/un.h>
#include <sys/unpcb.h>
#include <sys/domain.h>
#include <netinet/in.h>
#include <net/route.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>
#include <net/raw_cb.h>
#include <sys/stream.h>
#include <sys/mount.h>
#include <nfs/nfs.h>
#define	KERNEL_FILE
#include <sys/file.h>
#include <ufs/inode.h>
/*
 * The following header files need _KERNEL and KERNEL defined.  Some
 * #undef preparation is necessary.  Three of the header files that the
 * following header files include --  <kern/queue.h>, <kern/timer.h>, and
 * <machine/pmap.h> -- have local copies with duplicate typedef's
 * conditionally eliminated by the LSOF definition.  The local copies
 * are in the dec_a subdirectory of the lsof distribution.
 */
#define _KERNEL	1
#define KERNEL	1
#undef	MACRO_END
#undef	PIPSIZ
#undef	i_forw
#undef	i_gen
#undef	i_gid
#undef	i_lock
#undef	i_mode
#undef	i_nlink
#undef	i_rdev
#undef	i_size
#undef	i_uid
#include <kern/task.h>
#include <s5fs/s5inode.h>
#include <sys/procfs.h>
#include <sys/proc.h>
#include <sys/user.h>
#undef	u_comm
#define	u_comm	uu_comm
#include <sys/specdev.h>
#include <sys/vnode.h>
#include <nfs/rnode.h>
#include <ufs/mfsnode.h>
#include <vm/vm_anon.h>
#include <vm/u_mape_seg.h>
#include <vm/vm_vp.h>

/*
 * The spec_node is not defined in a distributed header file, but in
 * a kernel source file.
 */

struct spec_node {
	struct vnode *sn_vnode;
	struct vattr sn_vattr;
};

/*
 * The is an educated guess at an ADVFS/MSFS node.
 */

struct advfsnode {
	long	d1[19];
	ino_t	a_number;
	int	d2[5];
	long	a_size;
};

#define	COMP_P		void
#define DEVINCR		1024	/* device table malloc() increment */
#define	KA_T		off_t
#define	KMEM		"/dev/kmem"
#define MALLOC_P	char
#define FREE_P		MALLOC_P
#define MALLOC_S	size_t
#define N_UNIX		"/vmunix"
#define QSORT_P		char
#define	READLEN_T	int
#define STRNCPY_L	int
#define	U_SIZE		sizeof(struct user)


/*
 * Global storage definitions (including their structure definitions)
 */

struct clone {
	int dx;			/* index of device entry in Devtp[] */
	struct clone *next;	/* forward link */

};
extern struct clone *Clone;

extern dev_t Clonedev;

extern int Kmem;

struct l_vfs {
	struct mount *addr;		/* kernel address */
	fsid_t	fsid;			/* file system ID */
	short type;			/* type of file system */
	char *dir;			/* mounted directory */
	char *fsname;			/* file system name */
	struct l_vfs *next;		/* forward link */
};
extern struct l_vfs *Lvfs;

struct mounts {
        dev_t dev;              	/* st_dev */
        char *dir;              	/* directory */
        char *fsname;           	/* file system */
	ino_t inode;			/* st_ino */
	mode_t mode;			/* st_mode */
        struct mounts *next;    	/* forward link */
	dev_t rdev;			/* st_rdev */
};
extern struct mounts *Mtab;

extern struct mounts *Mtprocfs;

#define X_PROC          0
#define X_NPROC         1
#define X_VNMXP		2
#define X_CLONE		3

struct procfsid {
	int pid;			/* search PID */
	struct procfsid *next;		/* forward link */
};
extern struct procfsid *Procfsid;
extern int Procsrch;

struct sfile {
	char *aname;			/* argument file name */
	char *name;			/* file name (after readlink()) */
	char *devnm;			/* device name (optional) */
	dev_t dev;			/* device */
	u_short mode;			/* S_IFMT mode bits from stat() */
	int type;			/* file type: 0 = file system
				 	 *	      1 = regular file */
	ino_t i;			/* inode number */
	int f;				/* file found flag */
	struct sfile *next;		/* forward link */

};

extern int Vnmxp;

#endif	/* !DECOSF_LSOF_H */
