/*
 * ddev.c - SunOS (Solaris 1.x and 2.x) device support functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: ddev.c,v 1.6 94/05/11 18:18:45 abe Exp $";
#endif


#include "lsof.h"


/*
 * Local static values
 */

static int Devx = 0;			/* current Devtp[] index */


_PROTOTYPE(static void make_devtp,(struct stat *s, char *p, MALLOC_S nl));
_PROTOTYPE(static void readdev,(void));
_PROTOTYPE(static void stkdir,(char ***d, int *n, int *x, char *p));



/*
 * make_devtp() - make Devtp[] entry
 */

static void
make_devtp(s, p, nl)
	struct stat *s;			/* device stat() buffer */
	char *p;			/* device path name */
	MALLOC_S nl;			/* path name length (including '\0') */
{

/*
 * Make room for another Devtp[] entry.
 */
	if (Devx >= Ndev) {
		Ndev += DEVINCR;
		if (Devtp == NULL)
			Devtp = (struct dev *)malloc(
				(MALLOC_S)(sizeof(struct dev) * Ndev));
		else
			Devtp = (struct dev *)realloc((MALLOC_P *)Devtp,
				(MALLOC_S)(sizeof(struct dev) * Ndev));
		if (Devtp == NULL) {
			(void) fprintf(stderr, "%s: no space for devices\n",
				Pn);
			exit(1);
		}
	}
/*
 * Store the device number, inode number, and name in the Devtp[] entry.
 */
	Devtp[Devx].rdev = s->st_rdev;
	Devtp[Devx].inode = s->st_ino;
	if ((Devtp[Devx].name = (char *)malloc(nl)) == NULL) {
		(void) fprintf(stderr, "%s: no space for /dev/%s\n", Pn, p);
		exit(1);
	}
	(void) strcpy(Devtp[Devx].name, p);
	Devx++;
}


/*
 * printchdevname() - print character device name
 */

int
printchdevname(rdev, f)
	int rdev;			/* device */
	int f;				/* 1 = print trailing '\n' */
{
	struct clone *c;
	int low, hi, mid;
	struct dev *dp;

#if	defined(solaris)
	struct pseudo *p;
#endif

	if (!Sdev)
		readdev();
/*
 * Search for clone parent.
 */
	if (Lf->is_stream && Clone) {
		for (c = Clone; c; c = c->next) {
			if (major((dev_t)rdev) == minor(c->rdev)) {
				if (f)
					(void) puts(c->nm);
				else
					(void) fputs(c->nm, stdout);
				return(1);
			}
		}
	}
/*
 * Search device table for match.
 */
	low = mid = 0;
	hi = Ndev - 1;
        while (low <= hi) {
                mid = (low + hi) / 2;
		dp = Sdev[mid];
		if ((long)rdev < (long)dp->rdev)
			hi = mid - 1;
		else if ((long)rdev > (long)dp->rdev)
                        low = mid + 1;
                else {
			if (Lf->inode < 0L || (ino_t)Lf->inode == dp->inode) {
				if (f)
					(void) puts(dp->name);
				else
					(void) fputs(dp->name, stdout);
				return(1);
			}
			if (Lf->inode < (long)dp->inode)
				hi = mid - 1;
			else
				low = mid + 1;
		}
        }

#if	defined(solaris)
/*
 * Search for pseudo device match on major device only.
 */
	for (p = Pseudo; p; p = p->next) {
		if (major((dev_t)rdev) == major(p->rdev)) {
			if (f)
				(void) puts(p->nm);
			else
				(void) fputs(p->nm, stdout);
			return(1);
		}
	}
#endif	/* solaris */

	return(0);
}


#if	defined(solaris)
/*
 * read_clone() - read Solaris clone device information
 */

void
read_clone()
{
	struct clone *c;
	char *cn;
	DIR *dfp;
	struct dirent *dp;
	MALLOC_S nl;
	char path[MAXNAMLEN+1];
	struct pseudo *p;
	int pl;
	struct stat sb;
	STRNCPY_L dnamel;
/*
 * Open the /devices/pseudo directory.
 */
	(void) strcpy(path, "/devices/pseudo");
	if ((dfp = opendir(path)) == NULL) {
		if ( ! Fterse)
			(void) fprintf(stderr, "%s: can't open %s\n",
				Pn, path);
		return;
	}
	(void) strcat(path, "/");
	pl = strlen(path);
/*
 * Scan the directory.
 */
	for (dp = readdir(dfp); dp; dp = readdir(dfp)) {
		if (dp->d_ino == 0 || dp->d_name[0] == '.')
			continue;
	/*
	 * Form the full path name and lstat() it.
	 */
		dnamel = strlen(dp->d_name);
		if ((nl = pl + dnamel) >= sizeof(path)) {
		    if ( ! Fterse)
			(void) fprintf(stderr,
				"%s: /devices/pseudo entry name too long: %s\n",
				Pn, dp->d_name);
		    continue;
		}
		(void) strncpy(&path[pl], dp->d_name, dnamel);
		path[nl++] = '\0';
		if (lstat(path, &sb) != 0) {
			if ( ! Fterse)
				(void) fprintf(stderr,
					"%s: can't lstat %s: %s\n",
					Pn, path, strerror(errno));
			continue;
		}
	/*
	 * Skip subdirectories and all but character devices.
	 */
		if ((sb.st_mode & S_IFMT) == S_IFDIR
		||  (sb.st_mode & S_IFMT) != S_IFCHR)
			continue;
	/*
	 * Make Devtp[] entry.
	 */
		make_devtp(&sb, path, nl);
	 /*
	  * Create a clone structure entry for "clone*:" devices.
	  *
	  * Make special note of network clones -- icmp, ip, tcp, and udp.
	  */
		if (strncmp(&path[pl], "clone", 5) == 0) {
			if ((cn = strrchr(&path[pl], ':')) == NULL)
			    continue;
		/*
		 * Allocate a clone structure.
		 */
			if ((c = (struct clone *)malloc(sizeof(struct clone)))
			== NULL) {
			    (void) fprintf(stderr,
				"%s: no space for network clone device: %s\n",
				Pn, path);
			    exit(1);
			}
		/*
		 * Allocate space for the path name.
		 */
			if ((c->nm = (char *)malloc(nl)) == NULL) {
			    (void) fprintf(stderr,
				"%s: no space for network clone name: %s\n",
				Pn, path);
			    exit(1);
			}
		/*
		 * Save the path name and device number.
		 */
			(void) strcpy(c->nm, path);
			c->rdev = sb.st_rdev;
		/*
		 * Make special note of a network clone device.
		 */
			cn++;
			if (strcmp(cn, "icmp") == 0 || strcmp(cn, "ip")  == 0
			||  strcmp(cn, "tcp")  == 0 || strcmp(cn, "udp") == 0)
				c->n = cn - path;
			else
				c->n = 0;
		/*
		 * Link the new clone entry to the rest.
		 */
			c->next = Clone;
			Clone = c;
			continue;
		}
	/*
	 * Save pseudo device information.
	 */
		if (minor(sb.st_rdev) == 0) {

		/*
		 * Allocate space for the pseduo device entry.
		 */
		    if ((p = (struct pseudo *) malloc(sizeof(struct pseudo)))
		    == NULL) {
			(void) fprintf(stderr,
			    "%s: no space for pseudo device: %s\n",
			    Pn, path);
			exit(1);
		    }
		/*
		 * Allocate space for the path name.
		 */
		    if ((p->nm = (char *)malloc(nl)) == NULL) {
			(void) fprintf(stderr,
			    "%s: no space for pseudo name: %s\n", Pn, path);
			exit(1);
		    }
		/*
		 * Fill in the entry and link it to the rest.
		 */
		    (void) strcpy(p->nm, path);
		    p->rdev = sb.st_rdev;
		    p->next = Pseudo;
		    Pseudo = p;
		}
	}
	(void) closedir(dfp);
}
#endif	/* solaris */


/*
 * readdev() - read names, modes and device types of everything in /dev
 *	       or /device (Solaris)
 */

static void
readdev()
{
	struct clone *c;
	DIR *dfp;
	int dn = 0;
	struct dirent *dp;
	char **dstk = NULL;
	int dx = 0;
	int err = 0;
	int i;
	MALLOC_S nl;
	char path[MAXNAMLEN+1];
	int pl;
	struct stat sb;

#if	defined(solaris)
	char *cn;
	STRNCPY_L dnamel;
	struct pseudo *p;
#endif


#if	defined(solaris)
	(void) stkdir(&dstk, &dn, &dx, "/devices");
#else
	(void) stkdir(&dstk, &dn, &dx, "/dev");
#endif

/*
 * Unstack the next directory.
 */
	while (--dx >= 0) {
		(void) strcpy(path, dstk[dx]);
		if ((dfp = opendir(path)) == NULL) {
			if (Fdevmsg)
				(void) fprintf(stderr, "%s: can't open %s\n",
					Pn, path);
			continue;
		}
		(void) strcat(path, "/");
		pl = strlen(path);
		(void) free((FREE_P *)dstk[dx]);
		dstk[dx] = NULL;
	/*
	 * Scan the directory.
	 */
		for (dp = readdir(dfp); dp; dp = readdir(dfp)) {
			if (dp->d_ino == 0 || dp->d_name[0] == '.')
				continue;
		/*
		 * Form the full path name and get its status.
		 */

#if	defined(solaris)
			dnamel = strlen(dp->d_name);
			if ((nl = pl + dnamel) >= sizeof(path))
#else
			if ((nl = pl + dp->d_namlen) >= sizeof(path))
#endif	/* solaris */

			{
				(void) fprintf(stderr,
					"%s: /dev entry name too long: %s\n",
					Pn, dp->d_name);
				exit(1);
			}
#if	defined(solaris)
			(void) strncpy(&path[pl], dp->d_name, dnamel);
#else
			(void) strncpy(&path[pl], dp->d_name,
				(STRNCPY_L)dp->d_namlen);
#endif	/* solaris */

			path[nl++] = '\0';
			if (lstat(path, &sb) != 0) {
				(void) fprintf(stderr,
					"%s: can't lstat %s: %s\n",
					Pn, path, strerror(errno));
				err = 1;
				continue;
			}
		/*
		 * If it's a subdirectory, stack its name for later
		 * processing.
		 */
			if ((sb.st_mode & S_IFMT) == S_IFDIR) {

#if	defined(solaris)
			/*
			 * Skip Solaris /devices/pseudo sub-directory --
			 * it has been examined in read_clone().
			 */
				if (strcmp(path, "/devices/pseudo") == 0)
					continue;
#endif	/* solaris */

				(void) stkdir(&dstk, &dn, &dx, path);
				continue;
			}
		/*
		 * Skip all but character devices.
		 */
			if ((sb.st_mode & S_IFMT) != S_IFCHR)
				continue;
		/*
		 * Make Devtp[] entry.
		 */
			make_devtp(&sb, path, nl);

#if	!defined(solaris)
		/*
		 * Save information on non-Solaris clone devices.
		 */
			if (major(sb.st_rdev) == CLONEMAJ) {

			/*
			 * Allocate space for a clone entry.
			 */
			    if ((c = (struct clone *)malloc(
				sizeof(struct clone)))
			    == NULL) {
				(void) fprintf(stderr,
				    "%s: no space for clone device: %s\n",
				    Pn, path);
				exit(1);
			    }
			/*
			 * Allocate space for the path name.
			 */
			    if ((c->nm = (char *)malloc(nl)) == NULL) {
				(void) fprintf(stderr,
				    "%s: no space for clone name: %s\n",
				    Pn, path);
				exit(1);
			    }
			/*
			 * Fill in the clone entry and link it to the rest.
			 */
			    (void) strcpy(c->nm, path);
			    c->rdev = sb.st_rdev;
			    c->next = Clone;
		  	    Clone = c;
			}
#endif	/* !solaris */

		}
		(void) closedir(dfp);
	}
/*
 * Free any directory stack space.
 */
	if (dstk != NULL)
		(void) free((FREE_P *)dstk);
/*
 * Reduce the Devtp[] table to its minimum size.
 */
	if (Ndev > Devx) {
		Ndev = Devx;
		Devtp = (struct dev *)realloc((MALLOC_P *)Devtp,
		      (MALLOC_S)(sizeof(struct dev) * Ndev));
	}
	if (err)
		exit(1);
/*
 * Allocate sorting pointers and sort them by Devtp[] device number.
 */
	if ((Sdev = (struct dev **)malloc((MALLOC_S)(sizeof(struct dev *)
	    * Ndev)))
	== NULL) {
		(void) fprintf(stderr, "%s: no space for device pointers\n",
			Pn);
		exit(1);
	}
	for (i = 0; i < Ndev; i++) {
		Sdev[i] = &Devtp[i];
	}
	(void) qsort((QSORT_P *)Sdev,(size_t)Ndev,(size_t)sizeof(struct dev *),
		compdev);
}


/*
 * stkdir() - stack directory name
 */

static void
stkdir(d, n, x, p)
	char ***d;		/* array of directory pointers */
	int *n;			/* number of pointers */
	int *x;			/* current index */
	char *p;		/* directory path */
{
	if (*d == NULL) {

	/*
	 * Allocate first entry.
	 */
		if ((*d = (char **)malloc(sizeof(char *))) == NULL) {

stkdir_nospace:

			(void) fprintf(stderr,
				"%s: no space for directory stack at %s\n",
				Pn, p);
			exit(1);
		}
		*n = 1;
		*x = 0;
	} else if (*x >= *n) {

	/*
	 * Allocate additional space as required.
	 */
		*n += 1;
		if ((*d = (char **)realloc((MALLOC_P *)*d,
		          (MALLOC_S)(*n * sizeof(char *))))
		== NULL)
			goto stkdir_nospace;
	}
/*
 * Allocate space for the name, copy it there and put its pointer on the stack.
 */
	if (((*d)[*x] = (char *)malloc((MALLOC_S)(strlen(p) + 1))) == NULL) {
		(void) fprintf(stderr, "%s: no space for %s\n", Pn, p);
		exit(1);
	}
	(void) strcpy((*d)[*x], p);
	*x += 1;
}


/*
 * The find_ch_ino() function is obtained from ../common/fchi.frag.
 */
