/*
 * dmnt.c - SunOS (Solaris 1.x and 2.x) mount support functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: dmnt.c,v 1.3 94/05/10 08:45:22 abe Exp $";
#endif


#include "lsof.h"


/*
 * readmnt() - read mount table
 */

int
readmnt()
{
	int err = 0;
	char *dn = NULL;
	char *ln;
	FILE *mfp;
	struct mounts *mtp;
	struct stat sb;

#if	defined(solaris)
	struct mnttab me;
	struct mnttab *mp;
#else
	struct mntent *mp;
#endif

/*
 * Open access to the mount table and read mount table entries.
 */

#if	defined(solaris)
	if ((mfp = fopen(MNTTAB, "r")) == NULL) {
                (void) fprintf(stderr, "%s: can't access %s\n", Pn, MNTTAB);
                return(0);
        }
	for (mp = &me; getmntent(mfp, mp) == 0;)
#else
        if ((mfp = setmntent(MOUNTED, "r")) == NULL) {
                (void) fprintf(stderr, "%s: can't access %s\n", Pn, MOUNTED);
                return(0);
        }
        while ((mp = getmntent(mfp)) != NULL)
#endif	/* solaris */

	{

	/*
	 * Skip loop-back mounts, since they are aliases for legitimate file
	 * systems and there is no way to determine that a vnode refers to a
	 * loop-back alias.
	 *
	 * Also skip file systems of type MNTTYPE_IGNORE or with the option
	 * MNTOPT_IGNORE, since they are auto-mounted file systems whose
	 * real entries (if they are mounted) will be separately identified
	 * by getmntent().
	 */

#if	defined(solaris)
		if (strcmp(mp->mnt_fstype, MNTTYPE_LO) == 0
		||  hasmntopt(mp, MNTOPT_IGNORE))
#else
		if (strcmp(mp->mnt_type, MNTTYPE_LO) == 0
		||  strcmp(mp->mnt_type, MNTTYPE_IGNORE) == 0)
#endif

			continue;
	/*
	 * Interpolate a possible symbolic directory link.
	 */
		if (dn)
			(void) free((FREE_P *)dn);

#if	defined(solaris)
		if ((dn=(char *)malloc((MALLOC_S)(strlen(mp->mnt_mountp) + 1)))
#else
		if ((dn = (char *)malloc((MALLOC_S)(strlen(mp->mnt_dir) + 1)))
#endif
		== NULL) {
			err = 1;
			break;
		}

#if	defined(solaris)
		(void) strcpy(dn, mp->mnt_mountp);
#else
		(void) strcpy(dn, mp->mnt_dir);
#endif

		if ((ln = Readlink(dn)) == NULL) {
		    if ( ! Fterse){
			(void) fprintf(stderr,
			    "      Output information may be incomplete.\n");
		    }
		    err = 2;
		    continue;
		}
		if (ln != dn) {
			(void) free((FREE_P *)dn);
			dn = ln;
		}
	/*
	 * Stat() the directory.
	 */
                if (statsafely(dn, &sb)) {
		    if ( ! Fterse) {
			(void) fprintf(stderr,
			    "%s: WARNING: can't stat() %s file system %s\n",

#if	defined(solaris)
			    Pn, mp->mnt_fstype, mp->mnt_mountp);
#else
			    Pn, mp->mnt_type, mp->mnt_dir);
#endif

			(void) fprintf(stderr,
			    "      Output information may be incomplete.\n");
			}
		    err = 2;
		    continue;
		}
	/*
	 * Allocate and fill a local mount structure.
	 */
		if ((mtp = (struct mounts *)malloc(sizeof(struct mounts)))
		== NULL) {
			err = 1;
			break;
		}
		if ((mtp->fsname =

#if	defined(solaris)
			(char *)malloc((MALLOC_S)(strlen(mp->mnt_special)+1)))
#else
			(char *)malloc((MALLOC_S)(strlen(mp->mnt_fsname)+1)))
#endif

		== NULL) {
			err = 1;
			break;
		}

#if	defined(solaris)
		(void) strcpy(mtp->fsname, mp->mnt_special);
#else
		(void) strcpy(mtp->fsname, mp->mnt_fsname);
#endif

		mtp->dir = dn;
		dn = NULL;
		mtp->next = Mtab;
		mtp->dev = sb.st_dev;
		mtp->rdev = sb.st_rdev;
		mtp->inode = sb.st_ino;
		mtp->mode = sb.st_mode;
		Mtab = mtp;
        }

#if	defined(solaris)
	(void) fclose(mfp);
#else
        (void) endmntent(mfp);
#endif

	if (dn)
		(void) free((FREE_P *)dn);
/*
 * Handle errors.
 */
	switch (err) {
	case 1:
		(void) fprintf(stderr, "%s: no space for mount at %s (%s)\n",

#if	defined(solaris)
			Pn, mp->mnt_special, mp->mnt_mountp);
#else
			Pn, mp->mnt_fsname, mp->mnt_dir);
#endif

		return(0);
	case 2:
		if (Fmount)
			exit(1);
		break;
	}
	return(1);
}


/*
 * readvfs() - read vfs structure
 */

struct l_vfs *
readvfs(ka, la)
	struct vfs *ka;			/* vfs structure kernel address, if
					 * must be read from kernel */
	struct vfs *la;			/* local vfs structure address, non-
					 * NULL if already read from kernel */
{
	struct vfs *v, tv;
	struct l_vfs *vp;

#if	!defined(solaris)
	struct mount m;
# endif

	if (!ka && !la)
		return(NULL);
	for (vp = Lvfs; vp; vp = vp->next) {
		if (ka == vp->addr)
			return(vp);
	}
	if ((vp = (struct l_vfs *)malloc(sizeof(struct l_vfs))) == NULL) {
		(void) fprintf(stderr, "%s: PID %d, no space for vfs\n",
			Pn, Lp->pid);
		exit(1);
	}
	vp->dir = NULL;
	vp->fsname = NULL;
/*
 * Read vfs structure from kernel, if necessary.
 */
	if (la)
		v = la;
	else {
		v = &tv;
		if (kread((KA_T)ka, (char *)v, sizeof(tv))) {
			(void) free((FREE_P *)vp);
			return(NULL);
		}
	}
/*
 * Complete mount information.
 */

#if	defined(solaris)
	(void) completevfs(vp, v->vfs_dev);
#else
	if (Ntype == N_HSFS)
		(void) completevfs(vp, (int)(v->vfs_fsid.val[0] & 0xffff));
	else if (v->vfs_data
	     &&  kread((KA_T)v->vfs_data, (char *)&m, sizeof(m)) == 0)
		    (void) completevfs(vp, (int)m.m_dev);
#endif	/* solaris */

	vp->next = Lvfs;
	vp->addr = ka;
	Lvfs = vp;
	return(vp);
}


/*
 * The completevfs() function is obtained from ../common/cvfs.frag.
 */

/* CVFS_DEVSAVE is not defined. */
#define	USECVFS		1
