/*
 * dnode.c - SunOS (Solaris 1.x and 2.x) node reading functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: dnode.c,v 1.11 94/05/27 22:05:10 abe Exp $";
#endif


#include "lsof.h"


#if	defined (solaris)
_PROTOTYPE(static char isvlocked,(struct vnode *va));


/*
 * isvlocked() - is a vnode locked
 */

static char
isvlocked(va)
	struct vnode *va;		/* local vnode address */
{

	struct filock f;
	off_t l;

	if (va->v_filocks == NULL)
		return(' ');
	if (kread((KA_T)va->v_filocks, (char *)&f, sizeof(f)))
		return(' ');
	if (f.set.l_whence == 0 && f.set.l_start == 0 && f.set.l_len == 0)
		l = 1;
	else
		l = 0;
	switch (f.set.l_type & (F_RDLCK | F_WRLCK)) {

	case F_RDLCK:
		return((l) ? 'R' : 'r');
	case F_WRLCK:
		return((l) ? 'W' : 'w');
	case (F_RDLCK + F_WRLCK):
		return('*');
	}
	return(' ');
}
#endif	/* solaris */


/*
 * process_node() - process vnode
 */

void
process_node(va)
	caddr_t va;			/* vnode kernel space address */
{
	char dev_ch[32];
	dev_t dev;
	struct fifonode f;
	struct hsnode h;
	struct inode i;
	struct vfs kv;
	struct vfs *la = NULL;
	struct module_info mi;
	char mn[STRNML];
	int ni = 0;
	struct qinit qi;
	struct rnode r;
	struct vnode rv;
	struct snode s;
	struct stdata sd;
	struct queue sh;
	struct tmpnode t;
	char *tn;
	char *ty;
	enum vtype type;
	struct vnode v;
	struct l_vfs *vfs;

#if	defined(solaris)
	struct clone *c;
	struct dev_info di;
	char din[DINAMEL];
	struct filock fi;
	int fx;
	struct snode rs;
#endif

/*
 * Read the vnode.
 */
	if ( ! va) {
		enter_nm("no vnode address");
		return;
	}
	if (readvnode((caddr_t)va, &v)) {
                enter_nm(Namech);
                return;
        }

#if	defined(solaris)
/*
 * Check for a Solaris socket.
 */
	if (v.v_stream) {
		for (c = Clone; c; c = c->next) {
		    if (c->n && major(v.v_rdev) == minor(c->rdev)) {
			process_socket(v.v_stream, &c->nm[c->n]);
			return;
		    }
		}
	}
/*
 * Determine the Solaris vnode type.
 */
	if (v.v_vfsp && kread((KA_T)v.v_vfsp, (char *)&kv, sizeof(kv)) == 0) {
	/*
	 * Check the file system type.
	 */
		fx = kv.vfs_fstype;
		if (fx > 0 && fx <= Fsinfomax) {
			if (strcmp(Fsinfo[fx-1], "fifofs") == 0) {
				Ntype = N_FIFO;
			} else if (strcmp(Fsinfo[fx-1], "nfs") == 0) {
			    Ntype = N_NFS;
			    kv.vfs_dev = makedev((major_t)kv.vfs_fsid.val[1],
						 (minor_t)kv.vfs_fsid.val[0]);
			} else if (strcmp(Fsinfo[fx-1], "tmpfs") == 0) {
				Ntype = N_TMP;
			} else if (strcmp(Fsinfo[fx-1], "hsfs") == 0) {
				Ntype = N_HSFS;
				kv.vfs_dev = (long) kv.vfs_fsid.val[0];
			}
		}
	}
	if (Ntype == N_REGLR) {
		switch (v.v_type) {

		case VFIFO:
			Ntype = N_FIFO;
			break;
		}
		if (Ntype == N_REGLR && v.v_stream) {
			Ntype = N_STREAM;
			Lf->is_stream = 1;
		}
	}
/*
 * Determine the Solaris lock state.
 */
	Lf->lock = isvlocked(&v);
/*
 * Establish the local virtual file system structure.
 */
	if (v.v_vfsp == NULL)
		vfs = NULL;
	else if ((vfs = readvfs(v.v_vfsp, &kv)) == NULL) {
		(void) sprintf(Namech, "bad vfs for %#x at %#x", va, v.v_vfsp);
		enter_nm(Namech);
		return;
	}
#else	/* !solaris */

/*
 * Determine the non-Solaris vnode type.
 */
	if (Nl[X_NFS_OPS].n_value
	&& (unsigned long)v.v_op == Nl[X_NFS_OPS].n_value)
		Ntype = N_NFS;
	else if (Nl[X_TMP_OPS].n_value
	     && (unsigned long)v.v_op == Nl[X_TMP_OPS].n_value)
		Ntype = N_TMP;
	else if (Nl[X_HSFS_OPS].n_value
	     && (unsigned long)v.v_op == Nl[X_HSFS_OPS].n_value) {
	/*
	 * A Sun High Sierra vnode's virtual file system structure contains
	 * the major/minor device numbers in word 0 of its file system ID
	 * array.
	 */
		Ntype = N_HSFS;
		if (!v.v_vfsp
		||  kread((KA_T)v.v_vfsp, (char *)&kv, sizeof(kv))) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read vfs (%#x)",
				va, v.v_vfsp);
			enter_nm(Namech);
			return;
		}
		la = &kv;
	} else if (v.v_type == VFIFO)
		Ntype = N_FIFO;
	else if (v.v_type == VCHR && !v.v_vfsp && v.v_stream) {
		Ntype = N_STREAM;
		Lf->is_stream = 1;
	}
/*
 * Determine the non-Solaris lock type.
 */
	if (v.v_shlockc || v.v_exlockc) {
		if (v.v_shlockc && v.v_exlockc)
			Lf->lock = '*';
		else if (v.v_shlockc)
			Lf->lock = 'R';
		else
			Lf->lock = 'W';
	}
/*
 * Establish the local virtual file system structure.
 */
	if (v.v_vfsp == NULL)
		vfs = NULL;
	else if ((vfs = readvfs(v.v_vfsp, la)) == NULL) {
                (void) sprintf(Namech, "can't read vfs for %#x at %#x", va,
			v.v_vfsp);
                enter_nm(Namech);
		return;
	}
#endif	/* solaris */

/*
 * Read the fifonode, inode, rnode, snode or tmpnode.
 */
	switch (Ntype) {
	case N_NFS:
		if (!v.v_data || readrnode((caddr_t)v.v_data, &r)) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read rnode (%#x)",
				va, v.v_data);
			enter_nm(Namech);
			return;
		}
		break;
	case N_FIFO:

#if	defined(solaris)
	/*
	 * Solaris FIFO vnodes are linked to a fifonode.
	 * Non-pipe fifonodes are linked to a vnode thorough fn_realvp.
	 */
		if (!v.v_data || readfifonode(v.v_data, &f)) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read fifonode (%#x)",
				va, v.v_data);
			enter_nm(Namech);
			return;
		}
		if (f.fn_realvp) {
			if (readvnode((caddr_t)f.fn_realvp, &rv)) {
			    (void) sprintf(Namech,
				"fifonode at %#x: can't read real vnode (%#x)",
				v.v_data, f.fn_realvp);
			    enter_nm(Namech);
			    return;
			}
			if (!rv.v_data
			||  readinode((struct inode *)rv.v_data, &i)) {
				(void) sprintf(Namech,
				    "fifonode at %#x: can't read inode (%#x)",
				    v.v_data, rv.v_data);
				enter_nm(Namech);
				return;
			}
		}
		break;
#else	/* !solaris */

	/*
	 * A non-Solaris FIFO vnode is linked to a fifonode.  The s_realvp
	 * pointer of the snode contained in the fifonode points to a vnode
	 * that points to the inode.  (Whew!)
	 */
		if (!v.v_data || readfifonode(v.v_data, &f)) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read fifonode (%#x)",
				va, v.v_data);
			enter_nm(Namech);
			return;
		}
		if (f.fn_snode.s_realvp) {
			if (readvnode((caddr_t)f.fn_snode.s_realvp, &rv)) {
			    (void) sprintf(Namech,
				"fifonode at %#x: can't read real vnode (%#x)",
				v.v_data, f.fn_snode.s_realvp);
			    enter_nm(Namech);
			    return;
			}
			if (!rv.v_data
			||  readinode((struct inode *)rv.v_data, &i)) {
				(void) sprintf(Namech,
				    "fifonode at %#x: can't read inode (%#x)",
				    v.v_data, rv.v_data);
				enter_nm(Namech);
				return;
			}
		} else {
			enter_nm(Namech);
			return;
		}
		break;
#endif	/* solaris */

	case N_HSFS:

	/*
	 * A Sun High Sierra node contains the vnode.
	 * The device number is found in the virtual file system.
	 */
		if (!v.v_data || readhsnode((caddr_t)v.v_data, &h)) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read hsnode (%#x)",
				va, v.v_data);
			enter_nm(Namech);
			return;
		}
		break;
	case N_TMP:
		if (!v.v_data || readtnode((caddr_t)v.v_data, &t)) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read tnode (%#x)",
				va, v.v_data);
			enter_nm(Namech);
			return;
		}
		break;

	case N_STREAM:
		Lf->inode = -2L;
		(void) strcpy(Lf->iproto, "STR");
		if (!v.v_data || readsnode((caddr_t)v.v_data, &s)) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read snode (%#x)",
				va, v.v_data);
			enter_nm(Namech);
			return;
		}
		if (v.v_stream) {

		/*
		 * Follow the stream structure from its stdata pointer to the
		 * module information structure.  Then read the first STRNML-1
		 * characters of the module name.
		 */
			if (readstdata(v.v_stream, &sd) == 0
			&&  readsthead(sd.sd_wrq, &sh) == 0
			&&  readstqinit(sh.q_qinfo, &qi) == 0
			&&  readstmin(qi.qi_minfo,&mi) == 0
			&&  readstidnm(mi.mi_idname, mn, STRNML-1) == 0) {
				mn[STRNML-1] = '\0';
				if (mn[0])
					(void) strcpy(Namech, mn);
			}
		}
		break;
	case N_REGLR:
	default:
	/*
	 * VCHR vnodes point to an snode.  The snode's s_realvp usually
	 * points to a real vnode, which points to an inode (SunOS >= 4.1.3)
	 * or an rnode (SunOS < 4.1.3).
	 */
		if (v.v_type == VCHR) {
			if (!v.v_data || readsnode(v.v_data, &s)) {
				(void) sprintf(Namech,
					"vnode at %#x: can't read snode(%#x)",
					va, v.v_data);
				enter_nm(Namech);
				return;
			}
			if (s.s_realvp) {
			    if (readvnode((caddr_t)s.s_realvp, &rv)) {
				(void) sprintf(Namech,
				    "snode at %#x: can't read real vnode (%#x)",
				    v.v_data, s.s_realvp);
				enter_nm(Namech);
				return;
			    }

#if	!defined(solaris) && SUNOSV<40103
			    if (!rv.v_data
			    ||  readrnode((caddr_t)rv.v_data, &r))
#else
			    if (!rv.v_data
			    ||  readinode((struct inode *)rv.v_data, &i))
#endif
			    {
				(void) sprintf(Namech,
					"snode at %#x: can't read %snode (%#x)",
					v.v_data,

#if	!defined(solaris) && SUNOSV<40103
					"r",
#else
					"i",
#endif

					rv.v_data);
				enter_nm(Namech);
				return;
			    }
			}

#if	defined(solaris)
		/*
		 * When a Solaris snode has no s_realvp, but does have an
		 * s_commonvp, follow s_commonvp to a common vnode and then
		 * to a common snode.
		 *
		 * Follow the common snode to a dev_info structure and a
		 * device name.
		 */
			else if (s.s_commonvp) {
			    if (readvnode((caddr_t)s.s_commonvp, &rv)) {
				(void) sprintf(Namech,
				    "snode at %#x: can't read real vnode (%#x)",
				    v.v_data, s.s_commonvp);
				enter_nm(Namech);
				return;
			    }
			    if (!rv.v_data || readsnode(rv.v_data, &rs)) {
				(void) sprintf(Namech,
					"vnode at %#x: can't read snode(%#x)",
					s.s_commonvp, rv.v_data);
				enter_nm(Namech);
				return;
			    }
			    if (rs.s_dip) {
				if (kread((KA_T)rs.s_dip, (char *)&di,
				    sizeof(di)))
				{
				    (void) sprintf(Namech,
					"common snode at %#x: no dev info(%#x)",
					rv.v_data, rs.s_dip);
				    enter_nm(Namech);
				    return;
				}
				if (di.devi_name
				&& kread((KA_T)di.devi_name, din, DINAMEL-1)
				== 0) {
				    din[DINAMEL-1] = '\0';
				    (void) strcpy(Namech, din);
				    Lf->is_com = 1;
				}
			    }
			    ni = 1;
			}
# endif	/* solaris */

			else
				ni = 1;
			break;
		}
		if (!v.v_data || readinode((struct inode *)v.v_data, &i)) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read inode (%#x)",
				va, v.v_data);
			enter_nm(Namech);
			return;
		}
	}
/*
 * Get device and type for printing.
 */
	switch (Ntype) {
	case N_HSFS:

#if	defined(solaris)
		dev = kv.vfs_dev;
#else
		dev = kv.vfs_fsid.val[0] & 0xffff;
#endif

		break;
	case N_NFS:
		dev = r.r_attr.va_fsid;
		break;
	case N_STREAM:
		dev = s.s_dev;
		break;
	case N_TMP:
		dev = t.tn_attr.va_fsid;
		break;
	default:
		dev = (v.v_type == VCHR || ni) ? v.v_rdev : i.i_dev;
	}
	type = v.v_type;
	if (vfs && vfs->dir == NULL)
		(void) completevfs(vfs, (int)dev);
/*
 * Obtain the inode number.
 */
	switch (Ntype) {
	case N_HSFS:

#if	defined(solaris)
		Lf->inode = (long)h.hs_nodeid;
#else
		Lf->inode = (long)h.hs_dirent.ext_lbn;
#endif
		break;

	case N_NFS:
		Lf->inode = (long)r.r_attr.va_nodeid;
		break;
	case N_FIFO:

#if	defined(solaris)
		if ( ! f.fn_realvp) {
			Lf->inode = f.fn_ino;
			(void) sprintf(dev_ch, "0x%08x", v.v_data);
			enter_dev_ch(dev_ch);
			if (f.fn_flag & ISPIPE)
				(void) strcpy(Namech, "PIPE");
			if (f.fn_mate)
				(void) sprintf(endnm(), "->0x%08x", f.fn_mate);
			break;
		}
#endif

		/* fall through */

	case N_REGLR:

#if	!defined(solaris) && SUNOSV<40103 
		if (type == VCHR)
			Lf->inode = (long)r.r_attr.va_nodeid;
		else
#endif

		if (!ni)
			Lf->inode = (long)i.i_number;
		break;
	case N_TMP:
		Lf->inode = (long)t.tn_attr.va_nodeid;
		break;
	}
/*
 * Obtain the file size.
 */
	if (Foffset)
		Lf->off_def = 1;
	else {
		switch (Ntype) {

		case N_HSFS:
			Lf->sz = h.hs_dirent.ext_size;
			Lf->sz_def = 1;
			break;

#if	defined(solaris)
		case N_FIFO:
			if (!Fsize)
				Lf->off_def = 1;
			break;
#endif

		case N_NFS:

#if	defined(solaris)
			Lf->sz = r.r_size;
			Lf->sz_def = 1;
#else
			Lf->sz = r.r_attr.va_size;
			Lf->sz_def = 1;
#endif

			break;

		case N_REGLR:
			if (type == VREG || type == VDIR) {
				if (!ni) {
					Lf->sz = i.i_size;
					Lf->sz_def = 1;
				}
			} else if (type == VCHR && !Fsize)
				Lf->off_def = 1;
			break;
		case N_STREAM:
			if (!Fsize)
				Lf->off_def = 1;
			break;
		case N_TMP:
			Lf->sz = t.tn_attr.va_size;
			Lf->sz_def = 1;
		}
	}
/*
 * Record an NFS selection.
 */
	if (Ntype == N_NFS && Fnfs)
		Lf->sf |= SELNFS;
/*
 * Save the file system names.
 */
	if (vfs) {
		Lf->fsdir = vfs->dir;
		Lf->fsdev = vfs->fsname;
	}

#if	!defined(solaris)
/*
 * If this is the current working or root directory, and there is a path
 * name, use it.
 */
	if (*Cwd && strcmp(Lf->fd, CWD) == 0)
		(void) strcpy(Namech, Cwd);
	else if (*Rtd && strcmp(Lf->fd, RTD) == 0)
		(void) strcpy(Namech, Rtd);
#endif	/* !solaris */

/*
 * Format the vnode type, and possibly the device name.
 */
	switch (type) {

	case VNON:
		ty ="VNON";
		break;
	case VREG:
	case VDIR:
		ty = (type == VREG) ? "VREG" : "VDIR";
		Lf->dev_def = 1;
		Lf->dev = dev;
		break;
	case VBLK:
		ty = "VBLK";
		break;
	case VCHR:
		Lf->dev = dev;
		Lf->dev_def = 1;
		ty = "VCHR";
		if (Lf->is_stream == 0 && Lf->is_com == 0)
			Lf->is_chr_dev = 1;
		break;
	case VLNK:
		ty = "VLNK";
		break;

#if	defined(VSOCK)
	case VSOCK:
		ty = "SOCK";
		break;
#endif

	case VBAD:
		ty = "VBAD";
		break;
	case VFIFO:
		if (!Lf->dev_ch || Lf->dev_ch[0] == '\0') {
			Lf->dev = dev;
			Lf->dev_def = 1;
		}
		ty = "FIFO";
		break;
	default:
		if (type > 9999)
			(void) sprintf(Lf->type, "*%03d", type % 1000);
		else
			(void) sprintf(Lf->type, "%4d", type);
		(void) strcpy(Namech, "unknown type");
		ty = NULL;
	}
	if (ty)
		(void) strcpy(Lf->type, ty);

#if	defined(solaris)
/*
 * If this a Solaris common vnode/snode void some information.
 */
	if (Lf->is_com) {
		Lf->sz_def = 0;
		Lf->inode = -1L;
	}
#endif	/* solaris */

/*
 * If this is a VCHR file and it's missing an inode number, try to
 * supply one.
 */
	if (Lf->inode == -1L && type == VCHR && Lf->dev_def)
		find_ch_ino();
/*
 * Test for specified file.
 */
	if (Sfile && is_file_named(NULL, Ntype, type))
		Lf->sf |= SELNM;
/*
 * Enter name characters.
 */
	if (Namech[0])
		enter_nm(Namech);
}
