/*
 * main.c - common main function for lsof
 *
 * V. Abell, Purdue University Computing Center
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: main.c,v 1.4 94/05/21 09:42:43 abe Exp $";
#endif


#include "lsof.h"


/*
 * main() - main function for lsof
 */

main(argc, argv)
	int argc;
	char *argv[];
{
	int c, i;
	int err = 0;
	char options[64];
	struct sfile *sfp;
	struct lproc **slp;
	int sp = 1;

	if ((Pn = strrchr(argv[0], '/')) != NULL)
		Pn++;
	else
		Pn = argv[0];
/*
 * Common initialization.
 */
	Mypid = getpid();

#if	defined(HASSECURITY)
	Myuid = (uid_t)getuid();
#endif

	if ((Namech = (char *)malloc(MAXPATHLEN)) == NULL) {
		(void) fprintf(stderr, "%s: no space for name buffer\n", Pn);
		exit(1);
	}
/*
 * Create option mask.
 */
	(void) sprintf(options, "ac:dhHi:%slmnNoOp:PsS:tu:U",

#if	defined(HASKOPT)
			"k:"
#else
			""
#endif

			);
/*
 * Loop through options.
 */
	while ((c = getopt(argc, argv, options)) != EOF) {
		switch(c) {

		case 'a':
			Fand = 1;
			break;
		case 'c':
			Core = optarg;
			break;
		case 'd':
			Fdevmsg = 1;
			break;
		case 'h':
			Fhelp = 1;
			break;
		case 'H':
			Fhost = 0;
			break;
		case 'i':
			if (enter_network_address(optarg))
				err = 1;
			break;

#if	defined(HASKOPT)
		case 'k':
			Nmlst = optarg;
			break;
#endif

		case 'l':
			Futol = 0;
			break;

#if	defined(HASLOPT)
		case 'L':
			Flibdup = 0;
			break;
#endif

		case 'm':
			Fmount = 1;
			break;
		case 'n':
			Fnet = 1;
			break;
		case 'N':
			Fnfs = 1;
			break;
		case 'o':
			Foffset = 1;
			break;
		case 'O':
			Forder = 0;
			break;
		case 'p':
			if (enter_pid(optarg))
				err = 1;
			break;
		case 'P':
			Fport = 0;
			break;
		case 's':
			Fsize = 1;
			break;
		case 'S':
			if ((TmLimit = atoi(optarg)) < 2) {
			    TmLimit = TMLIMMIN;
			    (void) fprintf(stderr,
				"%s: -S time (%s) set to minimum: %d\n",
				Pn, optarg, TMLIMMIN);
			}
			break;
		case 't':
			Fterse = 1;
			break;
		case 'u':
			if (enter_uid(optarg))
				err = 1;
			break;
		case 'U':
			Funix = 1;
			break;
		case '?':
			err = 1;
			break;
		default:
			(void) fprintf(stderr, "%s: unknown option (%c)\n",
				Pn, c);
			err = 1;
		}
	}
	if (Fsize && Foffset) {
		(void) fprintf(stderr, "%s: -o and -s are mutually exclusive\n",
			Pn);
		err++;
	}
	if (err || Fhelp)
		usage(err);
/*
 * Compute the selection flags.
 */
	if (Fnet)
		Selflags |= SELNET;
	if (Fnfs)
		Selflags |= SELNFS;
	if (Funix)
		Selflags |= SELUNX;
	if (Npid)
		Selflags |= SELPID;
	if (Nuid)
		Selflags |= SELUID;
	if (Nwad)
		Selflags |= SELNA;
	if (optind < argc)
		Selflags |= SELNM;
	if (Selflags == 0) {
		if (Fand) {
			(void) fprintf(stderr,
				"%s: no select options to AND via -a\n", Pn);
			usage(1);
		}
		Selflags = SELALL;
	} else
		Selall = 0;
/*
 * Read the mount table unless only network or Unix domain socket files
 * are being displayed, or are part of an ANDed option specification.
 */
	if ((!Fand && (Selflags & ~(SELNA|SELNET|SELUNX)))
	||  ( Fand && (Selflags &  (SELNA|SELNET|SELUNX)) == 0)) {
		if (!readmnt())
			exit(1);
	}
/*
 * Process the file arguments.
 */
	if (optind < argc) {
		if (ck_file_arg(optind, argc, argv))
			usage(1);
	}
/*
 * Do dialect-specific initialization.
 */
	initialize();
/*
 * Gather information about processes.
 */
	gather_proc_info();
/*
 * Select the information to be displayed and sort it by Process ID,
 * unless sorting is inhibited.  In that case, the gather_proc_info()
 * function will have produced output as it became available.
 */
	if (Nlproc < 2)
		Forder = sp = 0;
	else if (Forder) {
		if ((slp = (struct lproc **)malloc(
			(MALLOC_S)(sizeof(struct lproc *) * Nlproc)))
		== NULL) {
		    (void) fprintf(stderr,
			"%s: WARNING -- no space for %d proc sort pointers\n",
			Pn, Nlproc);
		    sp = 0;
		}
		if (sp) {
			for (i = 0; i < Nlproc; i++) {
				slp[i] = &Lproc[i];
			}
			(void) qsort((QSORT_P *)slp, (size_t)Nlproc,
				     (size_t)sizeof(struct lproc *), comppid);
		}
	} else
		sp = 0;
	for (i = 0; i < Nlproc; i++) {
		Lp = sp ? slp[i] : &Lproc[i];
		if (Lp->pss)
			print_proc();
	}
/*
 * See if all requested information was displayed.  Return zero if it
 * was; one, if not.
 */
	for (sfp = Sfile; sfp; sfp = sfp->next) {
		if (sfp->f == 0)
			return(1);
	}
	return(0);
}
