/*
 * misc.c - common miscellaneous functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: misc.c,v 1.6 94/05/11 14:51:49 abe Exp $";
#endif


#include "lsof.h"

_PROTOTYPE(static int dolstat,(char *path, char *buf, int len));
_PROTOTYPE(static int doreadlink,(char *path, char *buf, int len));
_PROTOTYPE(static int doinchild,(int (*fn)(), char *fp, char *rbuf, int rbln));

#if	defined(HASINTSIGNAL)
_PROTOTYPE(static int handleint,(int sig));
#else
_PROTOTYPE(static void handleint,(int sig));
#endif


static jmp_buf Jmp_buf;		/* jump buffer */


/*
 * compdev() - compare Devtp[] entries
 */

int
compdev(a1, a2)
	COMP_P *a1, *a2;
{
	struct dev **p1 = (struct dev **)a1;
	struct dev **p2 = (struct dev **)a2;

	if ((long)((*p1)->rdev) < (long)((*p2)->rdev))
		return(-1);
	if ((long)((*p1)->rdev) > (long)((*p2)->rdev))
		return(1);
	if ((long)((*p1)->inode) < (long)((*p2)->inode))
		return(-1);
	if ((long)((*p1)->inode) > (long)((*p2)->inode))
		return(1);
	return(0);
}


/*
 * doinchild() -- do a function in a child process
 */

static int
doinchild(fn, fp, rbuf, rbln)
	int (*fn)();			/* function to perform */
	char *fp;			/* function parameter */
	char *rbuf;			/* response buffer */
	int rbln;			/* response buffer length */
{
	static int cpid;
	int en;
	int p[2], rv;
	static int wf;

/*
 * Set up to handle an alarm signal.
 *
 * Handle the signal.
 */
	cpid = wf = 0;
	if (setjmp(Jmp_buf)) {
	    (void) alarm(0);
	    (void) signal(SIGALRM, SIG_DFL);
	    if (wf) {
		if (cpid && ! Fterse)
		    (void) fprintf(stderr,
			"%s: WARNING -- child process %d may be hung.\n",
			Pn, cpid);
	    } else if (cpid) {
		(void) kill(cpid, SIGINT);
		(void) kill(cpid, SIGKILL);
		wf++;
		(void) signal(SIGALRM, handleint);
		(void) alarm(TmLimit);
		(void) wait(NULL);
		(void) alarm(0);
		(void) signal(SIGALRM, SIG_DFL);
	    }
	    errno = ETIMEDOUT;
	    return(-1);
	}
/*
 * Create a pipe from which to read the child function's response.
 */
	if (pipe(p) < 0) {
		(void) fprintf(stderr, "%s: can't open pipe: %s\n",
			Pn, strerror(errno));
		exit(1);
	}
/*
 * Fork a child to execute the function.
 */

	if ((cpid = fork()) == 0) {
		(void) close(p[0]);
		rv = fn(fp, rbuf, rbln);
		write(p[1], (char *)&rv, sizeof(rv));
		write(p[1], (char *)&errno, sizeof(errno));
		write(p[1], rbuf, rbln);
		(void) close(p[1]);
		_exit(0);
	}
	if (cpid < 0) {
		(void) fprintf(stderr, "%s: can't fork: %s\n",
			Pn, strerror(errno));
		exit(1);
	}
/*
 * Wait for the response in the parent.
 */
	(void) close(p[1]);
	(void) signal(SIGALRM, handleint);
	(void) alarm(TmLimit);
	if (read(p[0], (char *)&rv, sizeof(rv)) != sizeof(rv)
	||  read(p[0], (char *)&errno, sizeof(errno)) != sizeof(errno)
	||  read(p[0], rbuf, rbln) != rbln) {

	/*
	 * Cannot complete the read from the child.
	 */

		(void) kill(cpid, SIGINT);
		(void) kill(cpid, SIGKILL);
		(void) close(p[0]);
		wf++;
		(void) alarm(TmLimit);
		(void) wait(NULL);
		(void) alarm(0);
		(void) signal(SIGALRM, SIG_DFL);
		errno = EINTR;
		return(-1);
	}
/*
 * Function completed, response collected -- complete the operation.
 */
	en = errno;
	(void) alarm(0);
	(void) close(p[0]);
	wf++;
	(void) alarm(TmLimit);
	(void) wait(NULL);
	(void) alarm(0);
	(void) signal(SIGALRM, SIG_DFL);
	errno = en;
	return(rv);
}


/*
 * dolstat() - do an lstat() function
 */

static int
dolstat(path, rbuf, rbln)
	char *path;			/* path */
	char *rbuf;			/* response buffer */
	int rbln;			/* response buffer length */

/* ARGSUSED */

{
	return(lstat(path, (struct stat *)rbuf));
}


/*
 * doreadlink() -- do a readlink() function
 */

static int
doreadlink(path, rbuf, rbln)
	char *path;			/* path */
	char *rbuf;			/* response buffer */
	int rbln;			/* response buffer length */
{
	return(readlink(path, rbuf, rbln));
}


/*
 * enter_dev_ch() - enter device characters in file structure
 */

void
enter_dev_ch(m)
	char *m;
{
	MALLOC_S len;
	char *mp;

	if (m == NULL || *m == '\0')
		return;
	len = strlen(m) + 1;
	if ((mp = (char *)malloc(len)) == NULL) {
		(void) fprintf(stderr, "%s: no more dev_ch space at PID %d\n",
			Pn, Lp->pid);
		exit(1);
	}
	(void) strcpy(mp, m);
	if (Lf->dev_ch)
		(void) free((FREE_P *)Lf->dev_ch);
	Lf->dev_ch = mp;
}


/*
 * enter_nm() - enter name in local file structure
 */

void
enter_nm(m)
	char *m;
{
	MALLOC_S len;
	char *mp;

	if (m == NULL || *m == '\0')
		return;
	len = strlen(m) + 1;
	if ((mp = (char *)malloc(len)) == NULL) {
		(void) fprintf(stderr, "%s: no more nm space at PID %d\n",
			Pn, Lp->pid);
		exit(1);
	}
	(void) strcpy(mp, m);
	if (Lf->nm)
		(void) free((FREE_P *)Lf->nm);
	Lf->nm = mp;
}


/*
 * handleint() - handle an interrupt
 */

#if	defined(HASINTSIGNAL)
static int
#else
static void
#endif

handleint(sig)
	int sig;
{
	longjmp(Jmp_buf, 1);
}


/*
 * is_nw_addr() - is this network address selected?
 */

int
is_nw_addr(ia, p)
	struct in_addr *ia;		/* Internet address */
	int p;				/* port */
{
	struct nwad *n;
	unsigned char *u;

	if ((n = Nwad) == NULL)
		return(0);
	u = (unsigned char *)ia;
	for (; n; n = n->next) {
		if (n->proto) {
			if (strcasecmp(n->proto, Lf->iproto) != 0)
				continue;
		}
		if (n->a[0] != 0 || n->a[1] != 0
		||  n->a[2] != 0 || n->a[3] != 0) {
			if (u[3] != n->a[3] || u[2] != n->a[2]
			||  u[1] != n->a[1] || u[0] != n->a[0])
				continue;
		}
		if (n->port == -1 || p == n->port)
			return(1);
	}
	return(0);
}


/*
 * is_proc_excl() - is process excluded?
 */

int
is_proc_excl(pid, uid, pss, sf)
	int pid;			/* Process ID */
	UID_ARG uid;			/* User ID */
	short *pss;			/* process select state for lproc */
	short *sf;			/* select flags for lproc */
{
	int i;

	*pss = *sf = 0;

#if	defined(HASSECURITY)
/*
 * The process is excluded by virtue of the security option if it
 * isn't owned by the owner of this lsof process.
 */
	if (Myuid && Myuid != (uid_t)uid)
		return(1);
#endif

/*
 * If the listing of all processes is selected, then this one
 * is not excluded.
 */
	if (Selall) {
		*pss = PS_PRI;
		*sf = SELALL;
		return(0);
	}
/*
 * If the listing of processes has been specified by PID, see
 * if this one is specified.
 */
	if (Npid && (Selflags & SELPID)) {
		for (i = 0; i < Npid; i++) {
			if (Spid[i] == pid) {
				*pss = PS_PRI;
				*sf = SELPID;
				if (Selflags == SELPID)
					return(0);
				break;
			}
		}
		if (Selflags == SELPID && ! *sf)
			return(1);
	}
/*
 * If the listing of processes has been specified by UID, see
 * if the owner of this process has been specified.
 */
	if (Nuid && (Selflags & SELUID)) {
		for (i = 0; i < Nuid; i++) {
			if (Suid[i] == (uid_t)uid) {
				*pss = PS_PRI;
				*sf |= SELUID;
				if (Selflags == SELUID)
					return(0);
				break;
			}
		}
		if (Selflags == SELUID && (*sf & SELUID) == 0)
			return(1);
	}
/*
 * When neither the PID nor the UID is selected:
 *
 *	If list option ANDing of PIDs or UIDs is specified, the
 *	process is excluded;
 *
 *	Otherwise, it's not excluded by the tests of this function.
 */
	if ( ! *sf)
		return((Fand && (Selflags & (SELPID|SELUID))) ? 1 : 0);
/*
 * When either the PID or UID is selected and either the PID or UID
 * list option has been specified:
 *
 *	If list option ANDing has been specified, and the correct
 *	combination of PID and UID is selected, reply that the process
 *	is not excluded;
 * or
 *	If list option ANDing has not been specified, reply that the
 *	process is not excluded by the tests of this function.
 */
	if (Selflags & (SELPID|SELUID)) {
		if (Fand)
			return(((Selflags & (SELPID|SELUID)) != *sf) ? 1 : 0);
		return(0);
	}
/*
 * Finally, when neither the PID nor the UID is selected, and no PID
 * or UID list option has been specified:
 *
 *	If list option ANDing has been specified, this process is
 *	excluded;
 *
 *	Otherwise, it isn't excluded by the tests of this function.
 */
	return(Fand ? 1 : 0);
}


/*
 * Readlink() - read and interpret file system symbolic links
 */

char *
Readlink(arg)
	char *arg;			/* argument to be interpreted */
{
	char abuf[MAXPATHLEN];
	int alen;
	char *ap;
	char *argp1, *argp2;
	int len, llen, slen;
	char lbuf[MAXPATHLEN];
	char *s1;
	char tbuf[MAXPATHLEN];
/*
 * Evaluate each component of the argument for a symbolic link.
 */
	for (alen = 0, ap = abuf, argp1 = argp2 = arg; *argp2; argp1 = argp2 ) {
		for (argp2 = argp1 + 1; *argp2 && *argp2 != '/'; argp2++)
			;
		if ((len = argp2 - arg) >= MAXPATHLEN) {

path_too_long:
			if ( ! Fterse) {
				(void) fprintf(stderr,
					"%s: path too long: %s\n", Pn, arg);
			}
			return(NULL);
		}
		(void) strncpy(tbuf, arg, len);
		tbuf[len] = '\0';
	/*
	 * Dereference a symbolic link.
	 */
		if ((llen = doinchild(doreadlink, tbuf, lbuf, sizeof(lbuf) - 1))
		>= 0) {
		/*
		 * If the link is a new absolute path, replace
		 * the previous assembly with it.
		 */
			if (lbuf[0] == '/') {
				(void) strncpy(abuf, lbuf, llen);
				ap = &abuf[llen];
				*ap = '\0';
				alen = llen;
				continue;
			}
			lbuf[llen] = '\0';
			s1 = lbuf;
		} else {
			llen = argp2 - argp1;
			s1 = argp1;
		}
	/*
	 * Make sure two components are separated by a `/'.
	 *
	 * If the first component is not a link, don't force
	 * a leading '/'.
	 *
	 * If the first component is a link and the source of
	 * the link has a leading '/', force a leading '/'.
	 */
		if (*s1 == '/') {
			slen = 1;
		} else {
			if (alen > 0) {

			/*
			 * This is not the first component.
			 */
				if (abuf[alen - 1] == '/')
					slen = 1;
				else
					slen = 2;
			} else {

			/*
			 * This is the first component.
			 */
				if (s1 == lbuf && tbuf[0] == '/')
					slen = 2;
				else
					slen = 1;
			}
		}
	/*
	 * Add to the path assembly.
	 */
		if ((alen + llen + slen) >= sizeof(abuf))
			goto path_too_long;
		if (slen == 2)
			*ap++ = '/';
		(void) strncpy(ap, s1, llen);
		ap += llen;
		*ap = '\0';
		alen += (llen + slen - 1);
	}
/*
 * If the assembled path and argument are the same, return the argument.
 * Otherwise, allocate space for the assembled path and return it.
 */
	if (strcmp(arg, abuf) == 0)
		return(arg);
	if ((s1 = (char *)malloc((MALLOC_S)(alen + 1))) == NULL) {
		(void) fprintf(stderr,
			"%s: no space for symbolic link of %s\n", Pn, arg);
		exit(1);
	}
	(void) strcpy(s1, abuf);
	return(s1);
}


#if	defined(HASSTREAMS)
/*
 * readstdata() - read stream's stdata structure
 */

int
readstdata(addr, buf)
	struct stdata *addr;		/* stdata address in kernel*/
	struct stdata *buf;		/* buffer addess */
{
	if (addr == NULL
	||  kread((KA_T)addr, (char *)buf, sizeof(struct stdata))) {
		(void) sprintf(Namech, "no stream data in %#x\n", addr);
		return(1);
	}
	return(0);
}


/*
 * readsthead() - read stream head
 */

int
readsthead(addr, buf)
	struct queue *addr;		/* starting queue pointer in kernel */
	struct queue *buf;		/* buffer for queue head */
{
	struct queue *qp;

	if (addr == NULL) {
		(void) strcpy(Namech, "no stream queue head");
		return(1);
	}
	for (qp = addr; qp; qp = buf->q_next) {
		if (kread((KA_T)qp, (char *)buf, sizeof(struct queue))) {
			(void) sprintf(Namech, "bad stream queue link at %#x",
				qp);
			return(1);
		}
	}
	return(0);
}


/*
 * readstidnm() - read stream module ID name
 */

int
readstidnm(addr, buf, len)
	char *addr;			/* module ID name address in kernel */
	char *buf;			/* receiving buffer address */
	READLEN_T len;			/* buffer length */
{
	if (addr == NULL ||  kread((KA_T)addr, buf, len)) {
		(void) sprintf(Namech, "can't read module ID name from %#x",
			addr);
		return(1);
	}
	return(0);
}


/*
 * readstmin() - read stream's module info
 */

int
readstmin(addr, buf)
	struct module_info *addr;	/* module info address in kernel */
	struct module_info *buf;	/* receiving buffer address */
{
	if (addr == NULL
	||  kread((KA_T)addr, (char *)buf, sizeof(struct module_info))) {
		(void) sprintf(Namech, "can't read module info from %#x", addr);
		return(1);
	}
	return(0);
}


/*
 * readstqinit() - read stream's queue information structure
 */

int
readstqinit(addr, buf)
	struct qinit *addr;		/* queue info address in kernel */
	struct qinit *buf;		/* receiving buffer address */
{
	if (addr == NULL
	||  kread((KA_T)addr, (char *)buf, sizeof(struct qinit))) {
		(void) sprintf(Namech, "can't read queue info from %#x", addr);
		return(1);
	}
	return(0);
}
#endif	/* HASSTREAMS */


/*
 * statsafely() - stat path safely (i. e., with timeout)
 */

int
statsafely(path, buf)
	char *path;			/* file path */
	struct stat *buf;		/* stat buffer address */
{
	return(doinchild(dolstat, path, (char *)buf, sizeof(struct stat)));
}
