/*
 * print.c - common print support functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: print.c,v 1.3 94/05/04 17:01:53 abe Exp $";
#endif


#include "lsof.h"

#if	defined(EMPTY)
#undef	EMPTY
#endif

#include <utmp.h>

struct utmp dutmp;			/* dummy entry to get login name size */

#define HCINC	64			/* host cache size increase chunk */
#define LNML sizeof(dutmp.ut_name)
#define PCINC	512			/* port cache size increase chunk */

struct hostcache {
	unsigned char addr[4];
	char *name;
};

struct porttab {
	int port;
	char *name;
};

struct uidcache {
	uid_t u;
	char nm[LNML+1];
};

static int Npt[2] = { 0, 0};		/* TCP and UDP port entry counts */
static int Nuc = 0;			/* number of entries in UID cache */
static struct porttab *Pt[2] =		/* TCP and UDP port tables */
	{ NULL, NULL };
static int Pts = 0;			/* port tables initialization status */
static struct uidcache *Uc = NULL;	/* UID cache */
static unsigned Ucmask;			/* UID cache address mask */


/*
 * endnm() - locate end of Namech
 */

char *
endnm()
{
	char *s;

	for (s = Namech; *s; s++)
		;
	return(s);
}


/*
 * gethostnm() - get host name
 */

char *
gethostnm(ia)
	struct in_addr *ia;		/* pointer to Internet address */
{
	static struct hostcache *hc = NULL;
	static int hcx = 0;
	int i;
	struct hostent *he;
	static int nhc = 0;
	char *np;
	unsigned char *u;

	u = (unsigned char *)ia;
/*
 * Search cache.
 */
	for (i = 0; i < hcx; i++) {
		if (u[3] != hc[i].addr[3] || u[2] != hc[i].addr[2]
		||  u[1] != hc[i].addr[1] || u[0] != hc[i].addr[0])
			continue;
		return(hc[i].name);
	}
/*
 * Look up host name by address.
 */
	if ((he = gethostbyaddr((char *)ia, sizeof(struct in_addr), AF_INET))
	== NULL)
		return(NULL);
/*
 * Allocate space for name.
 */
	if ((np = he->h_name) != NULL) {
		if ((np = (char *)malloc(strlen(he->h_name) + 1)) == NULL) {
			(void) fprintf(stderr,
				"%s: no space for host name: %s\n",
				he->h_name);
			exit(1);
		}
		(void) strcpy(np, he->h_name);
	}
/*
 * Add address/name entry to cache.  Allocate cache space in HCINC chunks.
 */
	if (hcx >= nhc) {
		nhc += HCINC;
		if (hc == NULL)
			hc = (struct hostcache *)malloc(
				(MALLOC_S)(sizeof(struct hostcache) * nhc));
		else
			hc = (struct hostcache *)realloc((MALLOC_P *)hc,
				(MALLOC_S)(sizeof(struct hostcache) * nhc));
		if (hc == NULL) {
			(void) fprintf(stderr, "%s: no space for host cache\n",
				Pn);
			exit(1);
		}
	}
	hc[hcx].addr[0] = u[0];
	hc[hcx].addr[1] = u[1];
	hc[hcx].addr[2] = u[2];
	hc[hcx].addr[3] = u[3];
	hc[hcx++].name = np;
	return(np);
}


/*
 * iporttab() - initialize port table
 */

void
iporttab()
{
	int i[2];
	char *np;
	int p, ty;
	struct servent *se;

	if (Pts || !Fport)
		return;
	i[0] = i[1] = 0;
/*
 * Look up port name.
 */
	(void) setservent(1);
	while ((se = getservent()) != NULL) {
	    if (se->s_name == NULL)
		continue;

#if	defined(HASSWAPPORT)
	    p = htons(se->s_port);
#else
	    p = se->s_port;
#endif

	/*
	 * Determine protocol type -- TCP or UDP.  Ignore others.
	 */
	    if (strcasecmp(se->s_proto, "TCP") == 0)
		ty = 0;
	    else if (strcasecmp(se->s_proto, "UDP") == 0)
		ty = 1;
	    else
		continue;
	/*
	 * Make a copy of the service name.
	 */
	    if ((np = (char *)malloc(strlen(se->s_name) + 1)) == NULL) {
		(void) fprintf(stderr,
		    "%s: no space for %s port %d cache name: %s\n", Pn,
		    ty ? "UDP" : "TCP", p, se->s_name);
		(void) endservent();
		exit(1);
	    }
	    (void) strcpy(np, se->s_name);
	/*
	 * Allocate the port table in PCINC chunks.
	 */
	    if (i[ty] >= Npt[ty]) {
		Npt[ty] += PCINC;
		if (Pt[ty] == NULL)
		    Pt[ty] = (struct porttab *)malloc(
				(MALLOC_S)(sizeof(struct porttab) * Npt[ty]));
		else
		    Pt[ty] = (struct porttab *)realloc((MALLOC_P *)Pt[ty],
				(MALLOC_S)(sizeof(struct porttab) * Npt[ty]));
		if (Pt[ty] == NULL) {
		    (void) fprintf(stderr,
			"%s: no space for %s port table\n", Pn,
			ty ? "UDP" : "TCP");
		    (void) endservent();
		    exit(1);
		}
	    }
	/*
	 * Add the service name to the port table.
	 */
	    Pt[ty][i[ty]].port = p;
	    Pt[ty][i[ty]].name = np;
	    i[ty]++;
	}
/*
 * Reduce the port table sizes to the sizes of their contents.
 */
	(void) endservent();
	for (ty = 0; ty < 2; ty++) {
	    if (Pt[ty] == NULL)
		continue;
	    if ((Pt[ty] = (struct porttab *)realloc((MALLOC_P *)Pt[ty],
			  (MALLOC_S)(sizeof(struct porttab) * i[ty])))
	    == NULL) {
		(void) fprintf(stderr,
		    "%s: can't reduce %s port table size\n",
		    ty ? "UDP" : "TCP", Pn);
		exit(1);
	    }
	    Npt[ty] = i[ty];
	}
	Pts = 1;
}


/*
 * iuidcache() - initialize UID cache
 */

void
iuidcache(np)
	int np;				/* number of processes */
{
	int i;
	unsigned shft;

	if ( ! Futol)
		return;
/*
 * Set the first estimate of the cache size to 512, then adjust the size so
 * that it's a power of 2 and larger than the number of processes.
 */
	for (Nuc = 512, shft = 9; Nuc < np; shft++) {
		Nuc += Nuc;
	}
	Ucmask = (1 << shft) - 1;
/*
 * Allocate space for the cache and set all its entries empty.
 */
	if ((Uc = (struct uidcache *)malloc(Nuc * sizeof(struct uidcache)))
	== NULL) {
		(void) fprintf(stderr, "%s: no space for UID cache\n", Pn);
		exit(1);
	}
	for (i = 0; i < Nuc; i++) {
		Uc[i].nm[0] = '\0';
	}
}


/*
 * printinaddr() - print Internet address
 */

void
printinaddr(ia, p)
	struct in_addr *ia;		/* Internet address */
	int p;				/* port */
{
	char hbuf[32], pbuf[16];
	int i, ty;
	char *host = NULL;
	char *port = NULL;
	unsigned char *u;

	if ((Selflags & SELNA) && Nwad && is_nw_addr(ia, p))
		Lf->sf |= SELNA;
	if (ia->s_addr == INADDR_ANY)
		host = "*";
	else if (Fhost)
		host = gethostnm(ia);
	if (host == NULL) {
		u = (unsigned char *) ia;
		(void) sprintf(hbuf, "%u.%u.%u.%u", u[0], u[1], u[2], u[3]);
		host = hbuf;
	}
	if (p > 0) {
		if (Fport) {
			if (strcasecmp(Lf->iproto, "TCP") == 0)
				ty = 0;
			else if (strcasecmp(Lf->iproto, "UDP") == 0)
				ty = 1;
			else
				ty = 2;
			if (ty < 2) {
				if (! Pts)
					iporttab();
				for (i = 0; i < Npt[ty]; i++) {
					if (p == Pt[ty][i].port) {
						port = Pt[ty][i].name;
						break;
					}
				}
			}
		}
		if (port == NULL) {
			(void) sprintf(pbuf, "%d", p);
			port = pbuf;
		}
	} else if (p == 0)
		port = "0";
	(void) sprintf(endnm(), "%s%s%s", host,
		port ? ":"  : "",
		port ? port : "");
}


/*
 * printiproto() - print Internet protocol name
 */

void
printiproto(p)
	int p;				/* protocol number */
{
	int i;
	static int m = -1;
	char *s;

	switch(p) {

#if	defined(IPPROTO_TCP)
	case IPPROTO_TCP:
		s = "TCP";
		break;
#endif

#if	defined(IPPROTO_UDP)
	case IPPROTO_UDP:
		s = "UDP";
		break;
#endif

#if	defined(IPPROTO_IP)
	case IPPROTO_IP:
		s = "IP";
		break;
#endif

#if	defined(IPPROTO_ICMP)
	case IPPROTO_ICMP:
		s = "ICMP";
		break;
#endif

#if	defined(IPPROTO_IGMP)
	case IPPROTO_IGMP:
		s = "IGMP";
		break;
#endif

#if	defined(IPPROTO_GGP)
	case IPPROTO_GGP:
		s = "GGP";
		break;
#endif

#if	defined(IPPROTO_EGP)
	case IPPROTO_EGP:
		s = "EGP";
		break;
#endif

#if	defined(IPPROTO_PUP)
	case IPPROTO_PUP:
		s = "PUP";
		break;
#endif

#if	defined(IPPROTO_IDP)
	case IPPROTO_IDP:
		s = "IDP";
		break;
#endif

#if	defined(IPPROTO_ND)
	case IPPROTO_ND:
		s = "ND";
		break;
#endif

#if	defined(IPPROTO_RAW)
	case IPPROTO_RAW:
		s = "RAW";
		break;
#endif

#if	defined(IPPROTO_HELLO)
	case IPPROTO_HELLO:
		s = "HELLO";
		break;
#endif

#if	defined(IPPROTO_PXP)
	case IPPROTO_PXP:
		s = "PXP";
		break;
#endif

#if	defined(IPPROTO_RAWIP)
	case IPPROTO_RAWIP:
		s = "RAWIP";
		break;
#endif

#if	defined(IPPROTO_RAWIF)
	case IPPROTO_RAWIF:
		s = "RAWIF";
		break;
#endif

	default:
		s = NULL;
	}
	if (s)
		(void) sprintf(Lf->iproto, "%.*s", IPROTOL-1, s);
	else {	
		if (m < 0) {
			for (i = 0, m = 1; i < IPROTOL-2; i++)
				m *= 10;
		}
		if (m > p)
			(void) sprintf(Lf->iproto, "%d?", p);
		else
			(void) sprintf(Lf->iproto, "*%d?", p % (m/10));
	}
}


/*
 * printname() - print output name field
 */

void
printname()
{
	if (Lf->is_chr_dev && printchdevname(Lf->dev, 1)) 
		return;
	else if (Lf->is_com)
		fputs("COMMON: ", stdout);
	else if (Lf->is_stream) {
		if (printchdevname(Lf->dev, 0) && Lf->nm && Lf->nm[0])
			fputs("->", stdout);
	}
	if (Lf->nm && Lf->nm[0])
		(void) puts(Lf->nm);
	else if (Lf->fsdir || Lf->fsdev) {
		if (Lf->fsdir)
			(void) fputs(Lf->fsdir, stdout);
		if (Lf->fsdev) {
			if (Lf->fsdir)
				(void) fputs(" (", stdout);
			else
				(void) putchar('(');
			(void) fputs(Lf->fsdev, stdout);
			(void) puts(")");
		}
	} else
		(void) putchar('\n');
}


/*
 * printrawaddr() - print raw socket address
 */

void
printrawaddr(sa)
	struct sockaddr *sa;		/* socket address */
{
	(void) sprintf(endnm(), "%u/%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u",
		sa->sa_family,
		(unsigned char)sa->sa_data[0],
		(unsigned char)sa->sa_data[1],
		(unsigned char)sa->sa_data[2],
		(unsigned char)sa->sa_data[3],
		(unsigned char)sa->sa_data[4],
		(unsigned char)sa->sa_data[5],
		(unsigned char)sa->sa_data[6],
		(unsigned char)sa->sa_data[7],
		(unsigned char)sa->sa_data[8],
		(unsigned char)sa->sa_data[9],
		(unsigned char)sa->sa_data[10],
		(unsigned char)sa->sa_data[11],
		(unsigned char)sa->sa_data[12],
		(unsigned char)sa->sa_data[13]);
}


/*
 * printuid() - print User ID or login name
 */

char *
printuid(u)
	UID_ARG u;			/* User IDentification number */
{
	long i;
	int j = 0;
	int k;
	struct passwd *pw;
	static char user[USERPRTL+1];

	if (Futol && Uc) {

	/*
	 * If UID to login name conversion is in effect, and if there is
	 * a UID cache, search the cache.
	 */
		for (i = (long)u; j < Nuc; j++) {
			k = (int)(((i * 31415L) >> 7) & Ucmask);
			if (Uc[k].nm[0] == '\0') {

			/*
			 * Create a new cache entry.
			 */

#if	defined(HASPWSTAYOPEN)
				pw_stay_open();
#endif

				if ((pw = getpwuid(u)) == NULL) {
				    if ( ! Fterse) {
					(void) fprintf(stderr,
					    "%s: no pwd entry for UID %d\n",
					    Pn, u);
				    }
				    break;
				}
				(void) strncpy(Uc[k].nm, pw->pw_name, LNML);
				Uc[k].nm[LNML] = '\0';
				Uc[k].u = (uid_t)u;
			}
			if (Uc[k].u == (uid_t)u) {

			/*
			 * Use an existing cache entry.
			 */
				return(Uc[k].nm);
			}
		/*
		 * Reprobe after a cache collision.
		 */
			i = ((i + 1) >= Nuc) ? 0 : i + 1;
		}
	}
/*
 * Issue a warning if at the reprobe limit.
 */
	if (Nuc && j >= Nuc && ! Fterse) {
		(void) fprintf(stderr, "%s: at reprobe limit for UID %d\n",
			Pn, u);
	}
/*
 * Produce a numeric conversion of the UID.
 */
	(void) sprintf(user, "%*u", USERPRTL, u);
	return(user);
}
