/*
 * proc.c - common process and file structure functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: proc.c,v 1.4 94/05/04 17:02:13 abe Exp $";
#endif


#include "lsof.h"


/*
 * comppid() - compare PIDs
 */

int
comppid(a1, a2)
	COMP_P *a1, *a2;
{
	struct lproc **p1 = (struct lproc **)a1;
	struct lproc **p2 = (struct lproc **)a2;

	if ((*p1)->pid < (*p2)->pid)
		return(-1);
	if ((*p1)->pid > (*p2)->pid)
		return(1);
	return(0);
}


/*
 * alloc_lfile() - allocate local file structure space
 */

void
alloc_lfile(nm, num)
	char *nm;
	int num;
{
	if (Lf) {
/*
 * If reusing a previously allocated structure, release any allocated
 * space it was using.
 */
		if (Lf->dev_ch)
			(void) free((FREE_P *)Lf->dev_ch);
		if (Lf->nm)
			(void) free((FREE_P *)Lf->nm);
/*
 * Othwerise, allocate a new structure.
 */
	} else if ((Lf = (struct lfile *)malloc(sizeof(struct lfile)))
	== NULL) {
		(void) fprintf(stderr, "%s: no local file space at PID %d\n",
			Pn, Lp->pid);
		exit(1);
	}
/*
 * Initialize the structure.
 */
	Lf->access = Lf->lock = ' ';
	Lf->dev_def = Lf->is_chr_dev = Lf->is_com = Lf->is_stream
		    = Lf->off_def = Lf->sz_def = 0;
	Lf->off = 0;
	if (Lp->pss & PS_PRI)
		Lf->sf = Lp->sf;
	else
		Lf->sf = 0;
	Lf->iproto[0] = Lf->type[0] = '\0';
	if (nm) {
		(void) strncpy(Lf->fd, nm, FDLEN - 1);
		Lf->fd[FDLEN - 1] = '\0';
	} else if (num >= 0) {
		if (num < 10000)
			(void) sprintf(Lf->fd, "%4d", num);
		else
			(void) sprintf(Lf->fd, "*%03d", num % 1000);
	} else
		Lf->fd[0] = '\0';
	Lf->inode = -1L;
	Lf->dev_ch = Lf->fsdir = Lf->fsdev = Lf->nm = NULL;
	Lf->ch = -1;
	Lf->next = NULL;
	Ntype = N_REGLR;
	Namech[0] = '\0';
}


/*
 * alloc_lproc() - allocate local proc structure space
 */

void
alloc_lproc(pid, uid, cmd, pss, sf)
	int pid;			/* Process ID */
	UID_ARG uid;			/* User ID */
	char *cmd;			/* command */
	int pss;			/* process select state */
	int sf;				/* process select flags */
{
	static int sz;

	if (Lproc == NULL) {
		if ((Lproc = (struct lproc *)malloc(
		    (MALLOC_S) (LPROCINCR * sizeof(struct lproc))))
		== NULL) {
		    (void) fprintf(stderr,
			"%s: no malloc space for %d local proc structures\n",
			Pn, LPROCINCR);
		    exit(1);
		}
		sz = LPROCINCR;
	} else if ((Nlproc + 1) > sz) {
		sz += LPROCINCR;
		if ((Lproc = (struct lproc *)realloc((MALLOC_P *)Lproc,
		    (MALLOC_S) (sz * sizeof(struct lproc))))
		== NULL) {
		    (void) fprintf(stderr,
			"%s: no realloc space for %d local proc structures\n",
			Pn, sz);
		    exit(1);
		}
	}
	Lp = &Lproc[Nlproc++];
	(void) strncpy(Lp->cmd, cmd, CMDL-1);
	Lp->cmd[CMDL-1] = '\0';
	Lp->pid = pid;
	Lp->file = NULL;
	Lp->sf = (short)sf;
	Lp->pss = (short)pss;
	Lp->uid = (uid_t)uid;
}


/*
 * examine_lproc() - examine local process
 *
 * return: 1 = last process
 *
 * Note: the structures of an unselected process are not deallocated,
 *	 since they contain pointers to allocated and non-allocated
 *	 strings alike.
 */

int
examine_lproc()
{
	int sbp = 0;

/*
 * List the process if the process is selected an:
 *
 *	o  sorting is inhibited;
 *
 *	o  listing is limited to a single PID selection -- this one;
 *
 *	o  listing is selected by an ANDed option set (not all options)
 *	   that includes a single PID selection -- this one.
 */
	if ((Lp->sf & SELPID) && !Selall) {
		if ((Selflags == SELPID)
		||  (Fand && (Selflags & SELPID))) {
			sbp = 1;
			Npuns--;
		}
	}
	if (Lp->pss && (Forder == 0 || ((Npid == 1) && sbp))) {
		print_proc();
		Lp->pss = 0;
	}
/*
 * Deprecate an unselected (or listed) process.
 */
	if ( ! Lp->pss)
		Nlproc--;
/*
 * Indicate last-process if listing is limited to PID selections,
 * and all selected processes have been listed.
 */
	return((sbp && Npuns == 0) ? 1 : 0);
}


/*
 * is_file_sel() - is file selected?
 */

int
is_file_sel(lf)
	struct lfile *lf;		/* lfile structure pointer */
{
	if ( !lf || !lf->sf)
		return(0);
	if (Selall)
		return(1);
	if (Fand && ((lf->sf & Selflags) != Selflags))
		return(0);
	return(1);
}


/*
 * link_lfile() - link local file structures
 */

void
link_lfile()
{
	Lp->pss |= PS_SEC;
	if (Plf)
		Plf->next = Lf;
	else
		Lp->file = Lf;
	Plf = Lf;
	Lf = NULL;
}


/*
 * print_proc() - print process
 */

void
print_proc()
{
	int sel = 0;
/*
 * If nothing in the process has been selected, skip it.
 */
	if (!Lp->pss)
		return;
	if (Fterse) {

	/*
	 * The mode is terse and something in the process has been
	 * selected.  If options are being OR'd, print the PID;
	 * if AND'd, see if anything has been completed selected.
	 */
		if (Fand) {
			for (Lf = Lp->file; Lf; Lf = Lf->next) {
				if (is_file_sel(Lf)) {
					sel++;
					break;
				}
			}
		} else
			sel = 1;
		if (sel)
			(void) printf("%d\n", Lp->pid);
		return;
	}
/*
 * Print files.
 */
	for (Lf = Lp->file; Lf; Lf = Lf->next) {
		print_file();
	}
}
